import json
import numpy as np
import random
import uuid
from dexter.config.constants import Split
from dexter.data.datastructures.answer import AmbigNQAnswer
from dexter.data.datastructures.dataset import DprDataset
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.evidence import Evidence


from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams

from dexter.data.datastructures.sample import Sample
from dexter.data.loaders.BaseDataLoader import GenericDataLoader
import logging

logger = logging.getLogger(__name__)


class DprDataLoader(GenericDataLoader):
    def __init__(self, dataset: str, tokenizer="bert-base-uncased", config_path='test_config.ini', config=None,
                 split=Split.TRAIN,
                 batch_size=None):
        super().__init__(dataset, tokenizer, config_path, config, split, batch_size)

    def load_json(self, split=Split.TRAIN):
        dataset_path = f'{self.data_folder_path}/{self.config["File-Path"]["file"]}.json'
        with open(dataset_path, 'r') as fp:
            dataset = json.load(fp)
        return dataset

    def load_dataset(self, config: DenseHyperParams, split=Split.TRAIN, shuffle: bool = False):
        dataset = self.load_json(split)
        for data in dataset:
            _id = data['id'] if "id" in data.keys() else uuid.uuid4()
            question = Question(data["question"], None)

            evidences = {}

            evidences["positives"] = [Evidence(document["title"] + "" + document["text"], document["passage_id"]) for
                                      document in data["positive_ctxs"]]

            evidences["negatives"] = [Evidence(document["title"] + "" + document["text"], document["passage_id"]) for
                                      document in data["negative_ctxs"]]

            if "hard_negative_ctxs" in list(data.keys()) and len(data["hard_negative_ctxs"]) > 0:
                evidences["hard_negatives"] = [
                    Evidence(document["title"] + "" + document["text"], document["passage_id"]) for document in
                    data["hard_negative_ctxs"]]

            if shuffle:
                positive_ctxs = evidences["positives"]
                positive_ctx = positive_ctxs[np.random.choice(len(evidences["positive"]))]
                random.shuffle(evidences["negatives"])
                if "hard_negative_ctxs" in list(evidences.keys()):
                    random.shuffle(evidences["hard_negatives"])
            else:
                positive_ctx = [evidences["positives"][0]]
                evidences["positive_document"] = positive_ctx
            evidences["negatives"] = evidences["negatives"][:config.num_negative_samples]
            evidences["hard_negatives"] = evidences["negatives"][:config.num_negative_samples]
            for evidence in evidences["positive_document"]:
                self.raw_data.append(Sample(_id, question, AmbigNQAnswer(data["answers"]), evidence))
        ip_ids, ip_attention = self.tokenize_questions()
        evidence_ids, evidence_attention = self.tokenize_evidences()
        return DprDataset(ip_ids, ip_attention, evidence_ids, evidence_attention)



from dexter.config.constants import Split
from dexter.data.datastructures.answer import AmbigNQAnswer, Answer
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.sample import AmbigNQSample
from dexter.data.loaders.BaseDataLoader import GenericDataLoader
import tqdm

class AmbigQADataLoader(GenericDataLoader):
    '''Data loader class to load Datset from raw AmbigQA dataset.
    AmbigQA dataset consists of a questions each of which has clarified sub questions which again each have their own answers. 
    Additionally each question has multiple annotations
    
    Arguments:
    dataset (str): string containing the dataset alias
    tokenzier (str) : name of the tokenizer model. Set tokenizer as None, if only samples to be loaded but not tokenized and stored. This can help save time if only the raw dataset is needed.
    config_path (str) : path to the configuration file containing various parameters
    split (Split) : Split of the dataset to be loaded
    batch_size (int) : batch size to process the dataset.
    corpus: corpus containing all needed passages.
    
    '''
    def __init__(
        self,
        dataset: str,
        tokenizer="bert-base-uncased",
        config_path="test_config.ini",
        split=Split.TRAIN,
        batch_size=None,
        corpus=None
    ):
        self.corpus = corpus
        self.titles = [self.corpus[idx].title() for idx,_ in enumerate(self.corpus)]
        print(self.titles[100],self.corpus[100].title(),len(self.titles))
        super().__init__(dataset, tokenizer, config_path, split, batch_size)

    def load_raw_dataset(self, split=Split.TRAIN):
        dataset = self.load_json(split)
        for index, data in enumerate(tqdm.tqdm(dataset)):
            _id = data["id"]
            question = Question(data["question"], str(index))
            sample_answers = []
            for annotation in data["annotations"]:
                if annotation["type"] == "singleAnswer":
                    answers = [
                        [Answer(answer, None) for answer in annotation["answer"]]
                    ]
                elif annotation["type"] == "multipleQAs":
                    answers = [
                        [Answer(answer, None) for answer in pair["answer"]]
                        for pair in annotation["qaPairs"]
                    ]
                else:
                    raise TypeError("Unknown annotation type: ", annotation["type"])
                sample_answers.append(answers)
            for article in data["viewed_doc_titles"]:
                    if article in self.titles:
                        idx_lookup = list(self.titles).index(article)
                        evidence = Evidence(text= self.corpus[idx_lookup].text(),
                                              title=article,
                                              idx=idx_lookup)

                    self.raw_data.append(
                                AmbigNQSample(_id, question, AmbigNQAnswer(sample_answers), evidences=evidence)
                            )

    def tokenize_answers(self, MAX_LENGTH=20):
        # tokenize answers and make list for each tokenized answer
        decoder_ids = []
        decoder_masks = []
        for sample in self.raw_data:
            sample_op_ids = []
            sample_op_attention_masks = []
            for answer in sample.answer.flatten():
                tokenized_ans = self.tokenizer.tokenize(
                    answer,
                    pad_to_max_length="bart" in self.tokenizer_name,
                    max_length=MAX_LENGTH,
                )
                sample_op_ids.append(tokenized_ans["input_ids"])
                sample_op_attention_masks.append(tokenized_ans["attention_mask"])
            decoder_ids.append(sample_op_ids)
            decoder_masks.append(sample_op_attention_masks)
        return decoder_ids, decoder_masks


'''This file contains the hyperparameters to setup the Dense Passage Retriever'''

from dexter.data.datastructures.hyperparameters.base import BaseHyperParameters


class DenseHyperParams(BaseHyperParameters):
    def __init__(self, query_max_length: int=128,
                 query_encoder_path: str = "facebook/dpr-question_encoder-multiset-base",
                 document_encoder_path: str = "facebook/dpr-ctx_encoder-multiset-base",
                 learning_rate: float = 1e-5,
                 num_negative_samples: int = 5,
                 ann_search: str = "faiss_search",convert_to_tensor: bool = True, 
                 show_progress_bar: bool = True,convert_to_numpy: bool = True, batch_size: int = 64) -> None:
        super().__init__()

        self.query_max_length = query_max_length
        self.query_encoder_path = query_encoder_path
        self.document_encoder_path = document_encoder_path
        self.learning_rate = learning_rate
        self.num_negative_samples = num_negative_samples
        self.ann_search = ann_search
        self.convert_to_tensor = convert_to_tensor
        self.show_progress_bar = show_progress_bar
        self.batch_size = batch_size
        self.convert_to_numpy = convert_to_numpy

    def get_all_params(self):
        config = {
            "query_length" : self.query_max_length,
            "query_encoder_path": self.query_encoder_path,
            "document_encoder_path": self.document_encoder_path,
            "learning_rate": self.learning_rate,
            "num_negative_samples": self.num_negative_samples,
            "ann_search": self.ann_search,
            "convert_to_tensor": self.convert_to_tensor,
            "convert_to_numpy":self.convert_to_numpy,
            "batch_size": self.batch_size,
            "show_progress_bar": self.show_progress_bar
        }
        return config



import json
import os
from dexter.config.constants import Split
from dexter.data.datastructures.dataset import DprDataset
from dexter.data.datastructures.evidence import Evidence, TableEvidence
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.sample import Sample
from dexter.data.loaders.BaseDataLoader import GenericDataLoader
from dexter.config.constants import DataTypes
from dexter.config.constants import DataTypes


class TATQADataLoader(GenericDataLoader):
    '''Data loader class to load Datset from raw TATQA dataset.
    TATQA dataset consists of a questions each of which have a single answer and variopus table and text evidences.
    
    Arguments:
    dataset (str): string containing the dataset alias
    tokenzier (str) : name of the tokenizer model. Set tokenizer as None, if only samples to be loaded but not tokenized and stored. This can help save time if only the raw dataset is needed.
    config_path (str) : path to the configuration file containing various parameters
    split (Split) : Split of the dataset to be loaded
    batch_size (int) : batch size to process the dataset.
    corpus Dict[str,Evidence]: corpus containing all needed passages.    
    '''
    
    def __init__(self, dataset: str, tokenizer="bert-base-uncased", config_path='test_config.ini', split=Split.TRAIN,
                 batch_size=None, corpus=None):
        super().__init__(dataset, tokenizer, config_path, split, batch_size)
        self.corpus = corpus

    def load_raw_dataset(self, split=Split.TRAIN):
        dataset = self.load_json(split)
        for data in dataset:
            for question in data["questions"]:
                q = Question(question["question"],question["uid"])
                count = 0
                
                if(DataTypes.TABLE in question["answer_from"]):
                    rows = [list(row) for row in zip(*data["table"]["table"])]
                    columns = rows[0]
                    table = rows[1:]
                    self.raw_data.append(Sample(q.id()+"-"+str(count), q, question["answer"], TableEvidence(table=table,columns=columns,idx=data["table"]["uid"])))
                    count+=1
                if(DataTypes.TEXT in question["answer_from"]):
                    for paragraph_id in question["rel_paragraphs"]:
                        self.raw_data.append(Sample(q.id()+"-"+str(count), q, question["answer"], Evidence(data["paragraphs"][int(paragraph_id)-1]["text"],data["paragraphs"][int(paragraph_id)-1]["uid"])))
                        count+=1

    def load_tokenized(self):
        if self.tokenized_path and os.path.exists(self.tokenized_path):
            self.logger.info("Loading TATQA data from {}".format(self.tokenized_path))
            with open(self.tokenized_path, "r") as f:
                ip_ids, ip_attention, evidence_ids, evidence_attention = json.load(f)
        else:
            ip_ids, ip_attention = self.tokenize_questions()
            evidence_ids, evidence_attention = self.tokenize_evidences()
        return DprDataset(ip_ids, ip_attention, evidence_ids, evidence_attention)


import re
import string



class Metric:
    '''Base class for evaluating metric between predictions and targets'''
    def name(self):
        return "Not Defined"
    
    def normalize_answer(self,s):
        def remove_articles(text):
            return re.sub(r'\b(a|an|the)\b', ' ', text)

        def white_space_fix(text):
            return ' '.join(text.split())

        def remove_punc(text):
            exclude = set(string.punctuation)
            return ''.join(ch for ch in text if ch not in exclude)

        def lower(text):
            return text.lower()

        return white_space_fix(remove_articles(remove_punc(lower(s))))
    
    def evaluate(self,predictions,targets):
        pass


from flask import Flask, render_template, request
from functools import lru_cache
import math
import os
from dotenv import load_dotenv

from dexter.retriever.dense.ColBERT.colbert.infra import Run, RunConfig, ColBERTConfig
from colbert import Searcher

load_dotenv()

INDEX_NAME = os.getenv("INDEX_NAME")
INDEX_ROOT = os.getenv("INDEX_ROOT")
app = Flask(__name__)

searcher = Searcher(index=INDEX_NAME, index_root=INDEX_ROOT)
counter = {"api" : 0}

@lru_cache(maxsize=1000000)
def api_search_query(query, k):
    print(f"Query={query}")
    if k == None: k = 10
    k = min(int(k), 100)
    pids, ranks, scores = searcher.search(query, k=100)
    pids, ranks, scores = pids[:k], ranks[:k], scores[:k]
    passages = [searcher.collection[pid] for pid in pids]
    probs = [math.exp(score) for score in scores]
    probs = [prob / sum(probs) for prob in probs]
    topk = []
    for pid, rank, score, prob in zip(pids, ranks, scores, probs):
        text = searcher.collection[pid]            
        d = {'text': text, 'pid': pid, 'rank': rank, 'score': score, 'prob': prob}
        topk.append(d)
    topk = list(sorted(topk, key=lambda p: (-1 * p['score'], p['pid'])))
    return {"query" : query, "topk": topk}

@app.route("/api/search", methods=["GET"])
def api_search():
    if request.method == "GET":
        counter["api"] += 1
        print("API request count:", counter["api"])
        return api_search_query(request.args.get("query"), request.args.get("k"))
    else:
        return ('', 405)

if __name__ == "__main__":
    app.run("0.0.0.0", int(os.getenv("PORT")))




import heapq
from typing import Dict, List, Union, Tuple, Any
import numpy as np
import os
import joblib
from torch import Tensor
import torch
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
from sentence_transformers import SentenceTransformer
import logging
from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from dexter.retriever.BaseRetriever import BaseRetriver
from dexter.utils.metrics.SimilarityMatch import SimilarityMetric



class DenseFullSearch(BaseRetriver):

    def __init__(self,config=DenseHyperParams) -> None:
        super().__init__()
        self.config = config
        print("self.config.query_encoder_path",self.config.query_encoder_path)
        self.question_encoder = SentenceTransformer(self.config.query_encoder_path, device = "cuda")
        self.context_encoder = SentenceTransformer(self.config.document_encoder_path, device = "cuda")
        self.show_progress_bar = self.config.show_progress_bar
        self.convert_to_tensor = self.config.convert_to_tensor
        self.batch_size = self.config.batch_size
        self.sep="[SEP]"
        self.logger = logging.getLogger(__name__)
    
    def encode_queries(self, queries: List[Question], batch_size: int = 16, **kwargs) -> Union[List[Tensor], np.ndarray, Tensor]:
        queries = [query.text() for query in queries]
        return self.question_encoder.encode(queries, batch_size=batch_size,**kwargs)
    
    def encode_corpus(self, corpus: List[Evidence], **kwargs) -> Union[List[Tensor], np.ndarray, Tensor]:
        contexts = []
        for evidence in corpus:
            context = ""
            if evidence.title():
                context = (evidence.title() + self.sep + evidence.text()).strip()
            else:
                context = evidence.text().strip()
            contexts.append(context)
        return self.context_encoder.encode(contexts, batch_size=self.batch_size, **kwargs)

    def load_index_if_available(self)->Tuple[Any,bool]:
        if os.path.exists("indices/corpus/index"):
            corpus_embeddings = joblib.load("indices/corpus/index")
            index_present=True
        else:
            index_present = False
            corpus_embeddings=None
        return corpus_embeddings, index_present

    def retrieve_in_chunks(self,
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True,
                chunksize: int =200000,
                  **kwargs  ):
        corpus_ids = [doc.id() for doc in corpus]
        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size,show_progress_bar=self.show_progress_bar,convert_to_tensor=self.convert_to_tensor,**kwargs)  
        query_ids = [query.id() for query in queries]
        result_heaps = {qid: [] for qid in query_ids}  # Keep only the top-k docs for each query
        self.results = {qid: {} for qid in query_ids}
        batches = range(0, len(corpus), chunksize)
        for batch_num, corpus_start_idx in enumerate(batches):
            self.logger.info("Encoding Batch {}/{}...".format(batch_num+1, len(batches)))
            corpus_end_idx = min(corpus_start_idx + chunksize, len(corpus))

            # Encode chunk of corpus    
            sub_corpus_embeddings = self.encode_corpus(
                corpus[corpus_start_idx:corpus_end_idx],
                show_progress_bar=self.show_progress_bar, 
                convert_to_tensor = self.convert_to_tensor
                )

            # Compute similarites using either cosine-similarity or dot product
            cos_scores = score_function.evaluate(query_embeddings, sub_corpus_embeddings)
            cos_scores[torch.isnan(cos_scores)] = -1

            # Get top-k values
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
            
            for query_itr in range(len(query_embeddings)):
                query_id = query_ids[query_itr]                  
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                    if corpus_id != query_id:
                        if len(result_heaps[query_id]) < top_k:
                            # Push item on the heap
                            heapq.heappush(result_heaps[query_id], (score, corpus_id))
                        else:
                            # If item is larger than the smallest in the heap, push it on the heap then pop the smallest element
                            heapq.heappushpop(result_heaps[query_id], (score, corpus_id))

        for qid in result_heaps:
            for score, corpus_id in result_heaps[qid]:
                self.results[qid][corpus_id] = score
        return self.results
    def retrieve(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True, 
               chunk: bool = False,
               chunksize = None,
               **kwargs) -> Dict[str, Dict[str, float]]:

            
        self.logger.info("Encoding Queries...")

        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size,show_progress_bar=self.show_progress_bar,convert_to_tensor=self.convert_to_tensor,**kwargs)  
        self.logger.info("Encoding Corpus in batches... Warning: This might take a while!")

        if chunk:
            results = self.retrieve_in_chunks(corpus, 
                                              queries,top_k=top_k,
                                              score_function=score_function,return_sorted=return_sorted,
                                              chunksize=chunksize)
            return results
        #self.logger.info("Scoring Function: {} ({})".format(self.score_function_desc[score_function], score_function))
        embeddings, index_present = self.load_index_if_available()
        #TODO: Comment below for index usage
        #index_present = False
        if index_present:
            corpus_embeddings = embeddings
        else:
            corpus_embeddings = self.encode_corpus(corpus,show_progress_bar=self.show_progress_bar,convert_to_tensor=self.convert_to_tensor,**kwargs)
            #joblib.dump(corpus_embeddings,"indices/corpus/index")
        # Compute similarites using either cosine-similarity or dot product
        cos_scores = score_function.evaluate(query_embeddings,corpus_embeddings)
        # Get top-k values
        cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[0])), dim=1, largest=True, sorted=return_sorted)
        cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
        cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
        response = {}
        for idx, q in enumerate(queries):
            response[q.id()] = {}
            for index, c_idx in enumerate(cos_scores_top_k_idx[idx]):
                document_id = corpus[c_idx].id()
                response[q.id()][document_id] = float(cos_scores_top_k_values[idx][index])
        return response
                


from typing import List
import torch
import heapq
from sentence_transformers import SentenceTransformer
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from dexter.data.datastructures.question import Question

import logging

from dexter.retriever.dense.indexer.indexer import AnnSearch
from dexter.utils.metrics.SimilarityMatch import SimilarityMetric

logger = logging.getLogger(__name__)


class DprSentSearch():

    def __init__(self,
                 config: DenseHyperParams):
        self.query_encoder = SentenceTransformer(config.query_encoder_path, device='cuda')
        self.document_encoder = SentenceTransformer(
            config.document_encoder_path, device='cuda')
        self.documents = {}
        self.titles = {}
        self.data = []
        self.args = config
        self.ann_search = AnnSearch()
        self.ann_algo = None
        self.sep = "[SEP]"
        self.config = config
        self.logger = logging.getLogger(__name__)

    def get_passage_embeddings(self, passages:List[str] = None):
        #print("self.args.show_progress_bar",self.args.show_progress_bar)
        return self.document_encoder.encode(passages, convert_to_tensor=self.args.convert_to_tensor,
        convert_to_numpy=self.args.convert_to_numpy,
        show_progress_bar=self.args.show_progress_bar, batch_size=self.config.batch_size)

    def get_ann_algo(self, emb_dim, num_trees: int = None, metric: str = None):
        self.ann_algo = self.ann_search.get_ann_instance(
            self.args.ann_search, self.data,
            emb_dim, num_trees, metric)
        return self.ann_algo

    def create_index(self, corpus = None):
        for data in list(corpus):
            self.documents[data.id()] = data.text()
            self.titles[data.id()] = data.title()
                
        self.index_mapping = list(self.documents.keys())

        passages = [self.documents[idx]+"[SEP]"+self.titles[idx]
                         for idx in self.index_mapping]
        #print("passages",len(passages))
        index_exists = self.ann_algo.load_index_if_available()
        ##TODO: Uncomment below for index usage
        #index_exists = False
        if index_exists:
            logger.info(
                f'Index already exists. Loading {self.args.ann_search} index')
        else:
            passage_vectors = self.get_passage_embeddings(passages)
            assert len(passage_vectors)==len(self.index_mapping)
            self.ann_algo.create_index(passage_vectors)

    def retrieve_in_chunks(self,
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True,
                chunksize: int =400000,
                  **kwargs  ):
        queries_text = [query.text() for query in queries]
        corpus_ids = [doc.id() for doc in corpus]
        query_embeddings = self.query_encoder.encode(queries_text, 
            batch_size=self.config.batch_size,
            show_progress_bar=self.args.show_progress_bar,
            convert_to_tensor=self.args.convert_to_tensor)  
        query_ids = [query.id() for query in queries]
        result_heaps = {qid: [] for qid in query_ids}  # Keep only the top-k docs for each query
        self.results = {qid: {} for qid in query_ids}
        batches = range(0, len(corpus), chunksize)
        for batch_num, corpus_start_idx in enumerate(batches):
            self.logger.info("Encoding Batch {}/{}...".format(batch_num+1, len(batches)))
            corpus_end_idx = min(corpus_start_idx + chunksize, len(corpus))

            sub_corpus = corpus[corpus_start_idx:corpus_end_idx]
            # Encode chunk of corpus    
            contexts = []
            for evidence in sub_corpus:
                context = ""
                if evidence.title():
                    context = (evidence.title() + self.sep + evidence.text()).strip()
                else:
                    context = evidence.text().strip()
                contexts.append(context)
            sub_corpus_embeddings = self.get_passage_embeddings(
                contexts
                )

            # Compute similarites using either cosine-similarity or dot product
            cos_scores = score_function.evaluate(query_embeddings, sub_corpus_embeddings)
            cos_scores[torch.isnan(cos_scores)] = -1

            # Get top-k values
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
            
            for query_itr in range(len(query_embeddings)):
                query_id = query_ids[query_itr]                  
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                    if corpus_id != query_id:
                        if len(result_heaps[query_id]) < top_k:
                            # Push item on the heap
                            heapq.heappush(result_heaps[query_id], (score, corpus_id))
                        else:
                            # If item is larger than the smallest in the heap, push it on the heap then pop the smallest element
                            heapq.heappushpop(result_heaps[query_id], (score, corpus_id))

        for qid in result_heaps:
            for score, corpus_id in result_heaps[qid]:
                self.results[qid][corpus_id] = score
        return self.results

    def retrieve(self, queries:List[Question], top_k, chunk:bool = False, chunksize: int = 400000):
        query_vector = self.query_encoder.encode([query.text() for query in queries],convert_to_tensor=self.args.convert_to_tensor,show_progress_bar=self.args.show_progress_bar)
        top_neighbours = self.ann_algo.get_top_n_neighbours(
            query_vector, top_k)
        response = {}
        for idx,q in enumerate(queries):
            response[str(q.id())] = {}
            for index, c_idx in enumerate(top_neighbours["ids"][idx]):
                if(idx>=0):
                    response[str(q.id())][self.index_mapping[c_idx]] = float(top_neighbours["distances"][idx][index])
        return response



from typing import List

import pandas as pd
from dexter.config.constants import Separators

class Evidence:
    """ Data class to hold evidence/context for Question Answering

    Args:
        text : text of evidence passage
        title : title of evidence passage
        idx : index of evidence passage
    """
    def __init__(self, text: str, idx=None, title: str = None):
        self._text = text
        self._idx = idx
        self._title = title

    def text(self):
        return self._text

    def id(self):
        return self._idx

    def title(self):
        return self._title


class TableEvidence(Evidence):

    """ Data class to hold evidence/context for Question Answering in the format of a 2D table

    Args:
        table : 2D table of texts
        title: title of the evidence table
        columns : column names of the table
        idx : index of evidence table
    """
    def __init__(self, table: List[List],columns:List, idx=None, title: str = None):
        self.table = table
        self.columns = columns
        super().__init__(self.convert_to_text(), idx, title)

    def convert_to_text(self):
        return Separators.TABLE_ROW_SEP.join([Separators.TABLE_COL_SEP.join(row) for row in [self.columns]+self.table])

    def to_df(self):
        return pd.DataFrame(self.table, columns=self.columns)




import openai
from openai import OpenAI
import os

client = OpenAI(api_key=os.environ["OPENAI_KEY"])

class OpenAIEngine:
    def __init__(self, data,model_name: str, temperature: int, top_n: int):
        self.model_name = model_name
        self.temperature = temperature
        self.top_n = top_n
        self.data = data
    def get_completion(self, prompt: str) -> str:
        """_summary_

        Args:
            prompt (str): instruction with possibly in-context samples
        """        
        response = client.Completion.create(
            model="text-davinci-003",
            prompt="""You are a assistant that gives an answer along with the derivation of rationale in format Rationale: Answer:  to arrive at solution for questions mandatorily using information from both given table and text. In table columns are separated by | and rows by \n (newline). If you dont know the answer output UNKNOWN: \n
                      give best answer with accurate scale and precision for Question: """ + row["question"]+",Table: "+row["table"]+ "Text: "+row["text"]+"Output in format Rationale:, Answer:",
            temperature=self.temperature,
            max_tokens=2048,
            top_p=1.0,
            n = self.top_n,
            frequency_penalty=0.8,
            presence_penalty=0.6
            ) 
        return response['choices'][0]['text']     
    def get_chat_completion(self, user_prompt: str, system_prompt: str) -> str:
        """invokes chat completion endpoint to get chatpgt results

        Args:
            user_prompt (str): The user prompt with possibly in context samples
            system_prompt (str): system prompt that designates the responsibility of the system
        """        
        response = client.chat.completions.create(
            model="gpt-3.5-turbo",
            messages=[{"role":"system","content":system_prompt } ,
    {"role":"user","content": user_prompt,
    }] ,
            temperature=self.temperature,
            max_tokens=1000,
            top_p=1.0,
            n=self.top_n,
            frequency_penalty=0.8,
            presence_penalty=0.6,
            )
        return response.choices[0].message.content


from dexter.retriever.dense.ColBERT.baleen.condenser.condense import Condenser
from dexter.retriever.dense.ColBERT.baleen.hop_searcher import HopSearcher
from dexter.retriever.dense.ColBERT.baleen.engine import Baleen
from dexter.retriever.dense.ColBERT.colbert import Indexer
from dexter.retriever.dense.ColBERT.colbert.infra import Run, RunConfig, ColBERTConfig
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
import os
import logging
import tqdm
from typing import List,Dict

class BaleenRetriever():
    #Wrapper class for future extensions
    def __init__(self,config=ColBERTConfig,
                 checkpoint=None, condenser_path = None) -> None:
       # super().__init__(config)
        self.config = config
        self.indexer = Indexer(config=self.config,
                               checkpoint=checkpoint)
        self.sep="[SEP]"
        self.logger = logging.getLogger(__name__)
        self.checkpointL1 = os.path.join(condenser_path, 'hover.checkpoints-v1.0/condenserL1-v1.0.dnn')
        self.checkpointL2 = os.path.join(condenser_path, 'hover.checkpoints-v1.0/condenserL2-v1.0.dnn')
    def retrieve(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               return_sorted: bool = True, 
               index_name="colbert",
               **kwargs) -> Dict[str, Dict[str, float]]:
            corpus_texts = [(evidence.title() + self.sep + evidence.text()).strip() for evidence in corpus]
            queries_text = [question.text() for question in queries]
            with Run().context(RunConfig(nranks=1, experiment=index_name)):

                self.indexer.index(name=index_name, collection=corpus_texts, overwrite=True)

                self.indexer.get_index()
                searcher = HopSearcher(index=index_name)
                condenser = Condenser(checkpointL1=self.checkpointL1, 
                                      checkpointL2=self.checkpointL2,
                                    collectionX_path=collectionX_path,
                                    deviceL1='cuda:0', deviceL2='cuda:0')

                baleen = Baleen(collectionX_path, searcher, condenser)
                baleen.searcher.configure(nprobe=2, ncandidates=8192)
                result_qrels = {}
                for idx,query in tqdm.tqdm(enumerate(queries)):
                    result_qrels[str(query.id())] = {}
                    results = searcher.search(query.text(), k=top_k)
                    for passage_id, passage_rank, passage_score in zip(*results):
                        result_qrels[str(query.id())][str(corpus[passage_id].id())] = passage_score
                #results = searcher.search_all(queries_text,k=100)
            return result_qrels


from dexter.retriever.dense.ColBERT.colbert.modeling.tokenization.query_tokenization import *
from dexter.retriever.dense.ColBERT.colbert.modeling.tokenization.doc_tokenization import *
from dexter.retriever.dense.ColBERT.colbert.modeling.tokenization.utils import tensorize_triples



'''Base class for a retriever'''
class BaseRetriver:
    def __init__(self) -> None:
        pass

    def encode_queries(self)->None:
        pass

    def encode_context(self)->None:
        pass

    def train(self)->None:
        pass

    def retrieve(self)->None:
        pass

'''Retriever Factory to get retriever class given alias name'''

class RetrieverFactory():

    def get_retreiver(self,retriever_name:str,config_path:str):
        return None





import torch

from dexter.utils.metrics.MetricsBase import Metric

class SimilarityMetric(Metric):
    '''Base class for evaluating similarity between predictions and targets'''

    def name(self):
        return "None"  

    def normalize_embedding(self,embeddings):
        if not isinstance(embeddings, torch.Tensor):
            embeddings = torch.tensor(embeddings)
        if len(embeddings.shape) == 1:
            embeddings = embeddings.unsqueeze(0)
        return embeddings

    def evaluate(self,embeddings1,embeddings2):
        embeddings1 = self.normalize_embedding(embeddings1)
        embeddings2 = self.normalize_embedding(embeddings2)
        scores = self.score(embeddings1,embeddings2)
        scores[torch.isnan(scores)] = -1
        return scores

    def score(self,embeddings1,embeddings2):
        return None





class CosineSimilarity(SimilarityMetric):

    def name(self):
        return "Cosine Similarity"  
    
    def score(self,embeddings1,embeddings2):
        embeddings1_norm = torch.nn.functional.normalize(embeddings1, p=2, dim=1)
        embeddings2_norm = torch.nn.functional.normalize(embeddings2, p=2, dim=1)
        scores = torch.mm(embeddings1_norm, embeddings2_norm.transpose(0, 1))
        return scores

    
class DotScore(SimilarityMetric):
    def name(self):
        return "Dot Similarity"
    
    def score(self,embeddings1, embeddings2):
        scores  = torch.mm(embeddings1, embeddings2.transpose(0, 1))
        scores[torch.isnan(scores)] = -1
        return scores
    


import numpy as np
from dexter.utils.metrics.CoverExactMatch import CoverExactMatch


class AnswerF1:

    def __init__(self):
        self.em = CoverExactMatch()

    def get_f1(self, answers, predictions, is_equal=None):
        '''
        :answers: a list of list of strings
        :predictions: a list of strings
        '''
        assert len(answers)>0 and len(predictions)>0, (answers, predictions)
        occupied_answers = [False for _ in answers]
        occupied_predictions = [False for _ in predictions]
        for i, answer in enumerate(answers):
            for j, prediction in enumerate(predictions):
                if occupied_answers[i] or occupied_predictions[j]:
                    continue
                em = self.em.evaluate(answer, prediction)
                if em:
                    occupied_answers[i] = True
                    occupied_predictions[j] = True
        assert np.sum(occupied_answers)==np.sum(occupied_predictions)
        a, b = np.mean(occupied_answers), np.mean(occupied_predictions)
        if a+b==0:
            return 0
        return 2*a*b/(a+b)


import numpy as np
import re
import string
from dexter.utils.metrics.MetricsBase import Metric
'''This file contains function to calculate eact match between answers '''


class ExactMatch(Metric):
    def normalize_answer(self,s):

        def remove_articles(text):
            return re.sub(r'\b(a|an|the)\b', ' ', text)

        def white_space_fix(text):
            return ' '.join(text.split())

        def remove_punc(text):
            exclude = set(string.punctuation)
            return ''.join(ch for ch in text if ch not in exclude)

        def lower(text):
            return text.lower()

        return white_space_fix(remove_articles(remove_punc(lower(s))))
    def name(self):
        return "Exact Match"    
    def evaluate(self,answers1, answers2):
        if type(answers1)==list:
            if len(answers1)==0:
                return 0
            return np.max([self.evaluate(a, answers2) for a in answers1])
        if type(answers2)==list:
            if len(answers2)==0:
                return 0
            return np.max([self.evaluate(answers1, a) for a in answers2])
        return (self.normalize_answer(answers1) == self.normalize_answer(answers2))



import logging
from typing import List, Dict
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.evidence import Evidence

from dexter.re_ranker.CrossEncoder import CrossEncoder
class Reranker:
    
    def __init__(self, model_name, batch_size: int = 128, **kwargs):
        self.cross_encoder = CrossEncoder(model_name)
        self.batch_size = batch_size
        self.rerank_results = {}
        
    def rerank(self, 
               corpus: List[Evidence], 
               queries: List[Question],
               results: Dict[str, Dict[str, float]],
               top_k: int) -> Dict[str, Dict[str, float]]:
        
        sentence_pairs, pair_ids = [], []
        corpus_final = {}
        for evidence in corpus:
            corpus_final[str(evidence.id())] = {"title":evidence.title(),"text":evidence.text()}  

        queries_final = {}
        for index, query in enumerate(queries):
             queries_final[str(query.id())] = query.text()
        # print("query_ids**",query_ids)
        #queries = [query.text() for query in list(queries)]

        for query_id in results:
                for doc_id in results[query_id]:
                    pair_ids.append([query_id, doc_id])
                    corpus_text = (corpus_final[doc_id].get("title", "") + " " + corpus_final[doc_id].get("text", "")).strip()
                    sentence_pairs.append([queries_final[query_id], corpus_text])

        logging.info("Intitating Rerank Top-{}....".format(top_k))
        rerank_scores = [float(score) for score in self.cross_encoder.predict(sentence_pairs, batch_size=self.batch_size)]

        self.rerank_results = {query_id: {} for query_id in results}
        for pair, score in zip(pair_ids, rerank_scores):
            query_id, doc_id = pair[0], pair[1]
            self.rerank_results[str(query_id)][str(doc_id)] = score

        return self.rerank_results 


import logging
import numpy as np
import torch
import heapq
from numpy import ndarray
from torch import Tensor
import os

from tqdm.autonotebook import trange
from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from transformers import AutoModelForMaskedLM, AutoTokenizer
from sentence_transformers.util import batch_to_device
from typing import List, Dict, Tuple, Any, Union
import joblib
from dexter.retriever.BaseRetriever import BaseRetriver
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
from dexter.utils.metrics.SimilarityMatch import SimilarityMetric

logger = logging.getLogger(__name__)


class SPLADE(BaseRetriver):
    def __init__(self,config=DenseHyperParams):
        self.config = config
        #print("self.config.query_encoder_path",self.config.query_encoder_path)
        self.tokenizer = AutoTokenizer.from_pretrained(self.config.query_encoder_path)
        self.batch_size = self.config.batch_size
        self.sep="[SEP]"
        self.logger = logging.getLogger(__name__)
        self.model = SpladeNaver(self.config.query_encoder_path)
        self.model.eval()

    def load_index_if_available(self,data_name: str)->Tuple[Any,bool]:
        if os.path.exists("indices/corpus/splade_index_{}".format(data_name)):
            corpus_embeddings = joblib.load("indices/corpus/splade_index_{}".format(data_name))
            index_present=True
        else:
            index_present = False
            corpus_embeddings=None
        return corpus_embeddings, index_present
    # Write your own encoding query function (Returns: Query embeddings as numpy array)
    def encode_queries(self, queries: List[Question], **kwargs) -> np.ndarray:
        queries_text = [query.text() for query in queries]
        return self.model.encode_sentence_bert(self.tokenizer, queries_text, is_q=True, maxlen=self.config.query_max_length)

    # Write your own encoding corpus function (Returns: Document embeddings as numpy array)  out_features
    def encode_corpus(self, corpus: List[Evidence], **kwargs) -> np.ndarray:
        sentences = [(doc.title() + ' ' + doc.text()).strip() for doc in corpus]
        return self.model.encode_sentence_bert(self.tokenizer, sentences)
    def retrieve_in_chunks(self,
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True,
                chunksize: int =200000,
                  **kwargs  ):
        corpus_ids = [doc.id() for doc in corpus]
        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size)  
        query_ids = [query.id() for query in queries]
        result_heaps = {qid: [] for qid in query_ids}  # Keep only the top-k docs for each query
        self.results = {qid: {} for qid in query_ids}
        batches = range(0, len(corpus), chunksize)
        for batch_num, corpus_start_idx in enumerate(batches):
            self.logger.info("Encoding Batch {}/{}...".format(batch_num+1, len(batches)))
            corpus_end_idx = min(corpus_start_idx + chunksize, len(corpus))

            # Encode chunk of corpus    
            sub_corpus_embeddings = self.encode_corpus(
                corpus[corpus_start_idx:corpus_end_idx]
                )

            # Compute similarites using either cosine-similarity or dot product
            cos_scores = score_function.evaluate(query_embeddings, sub_corpus_embeddings)
            cos_scores[torch.isnan(cos_scores)] = -1

            # Get top-k values
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
            
            for query_itr in range(len(query_embeddings)):
                query_id = query_ids[query_itr]                  
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                    if corpus_id != query_id:
                        if len(result_heaps[query_id]) < top_k:
                            # Push item on the heap
                            heapq.heappush(result_heaps[query_id], (score, corpus_id))
                        else:
                            # If item is larger than the smallest in the heap, push it on the heap then pop the smallest element
                            heapq.heappushpop(result_heaps[query_id], (score, corpus_id))

        for qid in result_heaps:
            for score, corpus_id in result_heaps[qid]:
                self.results[qid][corpus_id] = score
        return self.results


    def retrieve(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               data_name: str,
               return_sorted: bool = True, 
               chunk: bool = False,
               chunksize = None,
               **kwargs) -> Dict[str, Dict[str, float]]:        
        self.logger.info("Encoding Queries...")

        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size)
        if chunk:
            results = self.retrieve_in_chunks(corpus, 
                                              queries,top_k=top_k,
                                              score_function=score_function,return_sorted=return_sorted,
                                              chunksize=chunksize)
            return results     

   
        self.logger.info("Encoding Corpus in batches... Warning: This might take a while!")
        #self.logger.info("Scoring Function: {} ({})".format(self.score_function_desc[score_function], score_function))
        embeddings, index_present = self.load_index_if_available(data_name)

        #TODO:Comment below for index usage
        #index_present = False
        if index_present:
            corpus_embeddings = embeddings
        else:
            corpus_embeddings = self.encode_corpus(corpus)
            if not os.path.exists("indices"):
                os.makedirs("indices")
            if not os.path.exists("indices/corpus"):
                os.makedirs("indices/corpus")
            joblib.dump(corpus_embeddings,"indices/corpus/splade_index_{}".format(data_name))
        

        # Compute similarites using either cosine-similarity or dot product
        cos_scores = score_function.evaluate(query_embeddings,corpus_embeddings)
        # Get top-k values
        cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
        cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
        cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
        response = {}
        for idx, q in enumerate(queries):
            response[q.id()] = {}
            for index, c_idx in enumerate(cos_scores_top_k_idx[idx]):
                response[q.id()][corpus[c_idx].id()] = float(cos_scores_top_k_values[idx][index])
        return response
                
# Chunks of this code has been taken from: https://github.com/naver/splade/blob/main/beir_evaluation/models.py
# For more details, please refer to SPLADE by Thibault Formal, Benjamin Piwowarski and Stéphane Clinchant (https://arxiv.org/abs/2107.05720)
class SpladeNaver(torch.nn.Module):
    def __init__(self, model_path):
        super().__init__()
        self.transformer = AutoModelForMaskedLM.from_pretrained(model_path)
        #self.transformer.cuda()
    def forward(self, **kwargs):
        out = self.transformer(**kwargs)["logits"]  # output (logits) of MLM head, shape (bs, pad_len, voc_size)
        return torch.max(torch.log(1 + torch.relu(out)) * kwargs["attention_mask"].unsqueeze(-1), dim=1).values

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """helper function to get the length for the input text. Text can be either
        a list of ints (which means a single text as input), or a tuple of list of ints
        (representing several text inputs to the model).
        """

        if isinstance(text, dict):  # {key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, '__len__'):  # Object has no len() method
            return 1
        elif len(text) == 0 or isinstance(text[0], int):  # Empty string or list of ints
            return len(text)
        else:
            return sum([len(t) for t in text])  # Sum of length of individual strings

    def encode_sentence_bert(self, tokenizer, sentences: Union[str, List[str], List[int]],
                             batch_size: int = 4,
                             show_progress_bar: bool = None,
                             output_value: str = 'sentence_embedding',
                             convert_to_numpy: bool = True,
                             convert_to_tensor: bool = False,
                             device: str = None,
                             normalize_embeddings: bool = False,
                             maxlen: int = 64,
                             is_q: bool = False) -> Union[List[Tensor], ndarray, Tensor]:
        """
        Computes sentence embeddings
        :param sentences: the sentences to embed
        :param batch_size: the batch size used for the computation
        :param show_progress_bar: Output a progress bar when encode sentences
        :param output_value:  Default sentence_embedding, to get sentence embeddings. Can be set to token_embeddings to get wordpiece token embeddings.
        :param convert_to_numpy: If true, the output is a list of numpy vectors. Else, it is a list of pytorch tensors.
        :param convert_to_tensor: If true, you get one large tensor as return. Overwrites any setting from convert_to_numpy
        :param device: Which torch.device to use for the computation
        :param normalize_embeddings: If set to true, returned vectors will have length 1. In that case, the faster dot-product (util.dot_score) instead of cosine similarity can be used.
        :return:
           By default, a list of tensors is returned. If convert_to_tensor, a stacked tensor is returned. If convert_to_numpy, a numpy matrix is returned.
        """
        self.eval()
        if show_progress_bar is None:
            show_progress_bar = True

        if convert_to_tensor:
            convert_to_numpy = False

        if output_value == 'token_embeddings':
            convert_to_tensor = False
            convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, '__len__'):
            # Cast an individual sentence to a list with length 1
            sentences = [sentences]
            input_was_string = True

        if device is None:
            device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

        self.to(device)

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
            sentences_batch = sentences_sorted[start_index:start_index + batch_size]
            # features = tokenizer(sentences_batch)
            # print(sentences_batch)
            features = tokenizer(sentences_batch,
                                 add_special_tokens=True,
                                 padding=True,  # pad to max sequence length in batch
                                 truncation=True,  # truncates to self.max_length
                                 return_attention_mask=True,
                                 return_tensors="pt")
            # print(features)
            features = batch_to_device(features, device)

            with torch.no_grad():
                out_features = self.forward(**features)
                if output_value == 'token_embeddings':
                    embeddings = []
                    for token_emb, attention in zip(out_features[output_value], out_features['attention_mask']):
                        last_mask_id = len(attention) - 1
                        while last_mask_id > 0 and attention[last_mask_id].item() == 0:
                            last_mask_id -= 1
                        embeddings.append(token_emb[0:last_mask_id + 1])
                else:  # Sentence embeddings
                    # embeddings = out_features[output_value]
                    embeddings = out_features
                    embeddings = embeddings.detach()
                    if normalize_embeddings:
                        embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
                    # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                    if convert_to_numpy:
                        embeddings = embeddings.cpu()
                all_embeddings.extend(embeddings)
        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]
        if convert_to_tensor:
            all_embeddings = torch.stack(all_embeddings)
        elif convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])
        if input_was_string:
            all_embeddings = all_embeddings[0]
        return all_embeddings


from transformers import AutoTokenizer


class Tokenizer:
    def __init__(self, tokenizer: str,prefix:str=None):
        self.prefix = prefix if prefix else ""
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)

    def tokenize(self, input_txt:str, **kwargs):
        return self.tokenizer(input_txt, **kwargs)

    def decode(self, idx):
        return self.tokenizer.decode(idx,
                                     skip_special_tokens=True,
                                     clean_up_tokenization_spaces=True).strip().replace(" - ", "-").replace(" : ", ":")



import configparser
from typing import Dict, List

from dexter.config.constants import Split
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
from dexter.data.loaders.BaseDataLoader import PassageDataLoader
from dexter.data.loaders.DataLoaderFactory import DataLoaderFactory

from dexter.data.loaders.Tokenizer import Tokenizer


class RetrieverDataset:
    '''Dataset class to load the data of the corresponding dataset provided for the evaluation of retrieval.
    Arguments
    dataset (str): alias of dataset
    passage_dataset (str): alias of corpus
    config_path (str) : path to the configuration file containing various parameters
    split (Split) : Split of the dataset to be loaded
    batch_size (int) : batch size to process the dataset.
     tokenzier (str) : name of the tokenizer model. Set tokenizer as None, if only samples to be loaded but not tokenized and stored. This can help save time if only the raw dataset is needed.
    '''
    def __init__(self, dataset:str,passage_dataset:str,config_path,split:Split, batch_size=32,tokenizer="bert-base-uncased"):
        self.split = split
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        self.tokenizer_name = tokenizer
        if(self.tokenizer_name):
            self.tokenizer = Tokenizer(self.tokenizer_name)
        else:
            self.tokenizer = None
        self.passage_dataloader = PassageDataLoader(passage_dataset,None,self.tokenizer_name,config_path)
        base_dataset = DataLoaderFactory().create_dataloader(dataset, config_path=config_path, split=self.split, batch_size=batch_size,tokenizer=self.tokenizer_name,
                                                             corpus=self.passage_dataloader.raw_data)
        self.base_dataset = base_dataset          
    
    
    def qrels(self)->(List[Question],Dict,List[Evidence]):
        qrels = {}
        queries = []
        corpus = self.passage_dataloader.raw_data
        for sample in self.base_dataset.raw_data:
            if str(sample.question.id()) not in list(qrels.keys()):
                qrels[str(sample.question.id())] = {}
                if(sample.question not in queries):
                    queries.append(sample.question)
            evidence = sample.evidences
            #print("str(sample.idx)",str(sample.idx),str(evidence.id()),qrels[str(sample.idx)])
            qrels[sample.question.id()][str(evidence.id())] = 1
        return queries,qrels,corpus


from typing import Optional

from dexter.data.datastructures.answer import Answer, AmbigNQAnswer
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question


class Sample:
    """
    A base class to hold one datapoint/sample with a question ans its corresponding answer

    Attributes:
        question (Question): question of the sample.
        answer (Answer): answer of the given question.
        evidenc (Answer): Optional context/evidence for the given question.
        _idx (int): The ID of the answer.
    """
    def __init__(self, idx, question: Question, answer: Answer, evidences: Optional[Evidence] = None):
        self.question = question
        self.evidences = evidences
        self.answer = answer
        self.idx = idx


class AmbigNQSample:
    """
    A base class to hold one datapoint/sample with a question ans its corresponding answer for the ambiguous case.

    Attributes:
        question (Question): question of the sample.
        answer (AmbigNQAnswer): answer of the given question.
        evidence (Answer): Optional context/evidence for the given question.
        _idx (int): The ID of the answer.
    """
    def __init__(self, idx, question: Question, answers: AmbigNQAnswer, evidences: Optional[Evidence] = None):
        self.question = question
        self.evidences = evidences
        self.answer = answers
        self.idx = idx



"""
This file holds the data loader classes for various stages in Complex Question Answering.
"""

import configparser
import copy
import gzip
import json
import os
from typing import List
import zipfile

from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from tqdm import tqdm

from dexter.config.constants import DataTypes, Separators, Split
from dexter.data.datastructures.answer import Answer
from dexter.data.datastructures.dataset import PassageDataset, QADataset
from dexter.data.datastructures.evidence import Evidence, TableEvidence
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.sample import Sample
from dexter.data.loaders.Tokenizer import Tokenizer


class GenericDataLoader(DataLoader):
    """
    Generic dataloader class to load raw datasets containing question and anwer pairs.

    Arguments:
    dataset (str): string containing the dataset alias
    tokenzier (str) : name of the tokenizer model.  Set tokenizer as None, if only samples to be loaded but not tokenized and stored. This can help save time if only the raw dataset is needed.
    config_path (str) : path to the configuration file containing various parameters
    split (Split) : Split of the dataset to be loaded
    prefix (str) : prefix string to concatenate question and answers with.
    """
    def __init__(
        self,
        dataset: str,
        tokenizer: str = "bert-base-uncased",
        config_path: str = "config.ini",
        split: Split = Split.TRAIN,
        prefix:str = None,
        batch_size: int = None,
    ):
        self.raw_data:List[Sample] = []
        self.meta = {}
        self.tokenizer_name = tokenizer
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        self.is_training = split == Split.TRAIN
        print(self.config["Data-Path"].keys())

        #Loads the poath to the raw dataset
        datapath = self.config["Data-Path"][dataset]

        #loads the path of pre tokenized data if saved before already.
        tokenized_path = f"{dataset}_{tokenizer}_tokenized"
        self.prefix = prefix if prefix else ""
        self.tokenized_path = (
            self.config["Data-Path"][tokenized_path]
            if tokenized_path in self.config["Data-Path"].keys()
            else None
        )
        if datapath.endswith("zip"):
            self.extract_zip_to_temp(dataset, self.config["Data-Path"][dataset])
            self.data_folder_path = dataset + "-" + "temp"
        else:
            self.data_folder_path = self.config["Data-Path"][dataset]

        self.load_raw_dataset(split)
        #If no tokenizer given, only load raw dataset
        if(self.tokenizer_name):
            self.tokenizer = Tokenizer(self.tokenizer_name)
            self.dataset = self.load_tokenized()

            if self.is_training:
                sampler = RandomSampler(self.dataset)
            else:
                sampler = SequentialSampler(self.dataset)
            print("Dataset loaded of length", len(self.dataset))
            super(GenericDataLoader, self).__init__(
                self.dataset, sampler=sampler, batch_size=batch_size
            )

    @staticmethod
    def extract_zip_to_temp(dataset: str, zip_file_path):
        with zipfile.ZipFile(zip_file_path, "r") as zip_ref:
            zip_ref.extractall(dataset + "-" + "temp")

    def load_json(self, split=Split.TRAIN):
        dataset_path = f"{self.data_folder_path}/{split}.json"
        with open(dataset_path, "r") as fp:
            dataset = json.load(fp)
        return dataset

    def load_tokenized(self):
        '''Loads the tokenized QA Dataset consisting of question answer samples'''
        if self.tokenized_path and os.path.exists(self.tokenized_path):
            self.logger.info("Loading DPR data from {}".format(self.tokenized_path))
            with open(self.tokenized_path, "r") as f:
                ip_ids, ip_attention, op_ids, op_attention = json.load(f)
        else:
            ip_ids, ip_attention = self.tokenize_questions()
            op_ids, op_attention = self.tokenize_answers()
        return QADataset(ip_ids, ip_attention, op_ids, op_attention, self.is_training)

    def load_raw_dataset(self, split=Split.TRAIN):
        '''Loads the raw question answer pairs and creates samples and stores as list'''
        dataset = self.load_json(split)
        for data in dataset:
            self.raw_data.append(
                Sample(data["id"], Question(data["question"]), Answer(data["answer"])))

    def tokenize_questions(self, MAX_LENGTH=32):
        '''Loads the questions in the dataset and returns the input ids and attention masks'''
        op = self.tokenizer.tokenize(
            [self.tokenizer.prefix+data.question.text() for data in self.raw_data],
            padding="max_length",
            max_length=MAX_LENGTH,
            truncation=True,
        )
        return op["input_ids"], op["attention_mask"]
    
    def tokenize_evidences(self):
        '''Loads the evidence passages associated to the correposding questions of the datase and returns the input ids and attention masks'''
        op = self.tokenizer.tokenize([data.evidences.text() for data in self.raw_data], padding=True, truncation=True,
                                     return_tensors="pt")
        return op['input_ids'], op['attention_mask']
        

    def tokenize_answers(self):
        '''Loads the answers associated to the correposding questions of the datase and returns the input ids and attention masks'''
        return self.tokenizer.tokenize([self.tokenizer.prefix+data.answer.text() for data in self.raw_data])

class PassageDataLoader(DataLoader):
    """
    Dataloader class to load the full corpus cotaining all passages.

    Arguments:
    dataset (str): string containing the dataset alias
    subset_ids List : list of passage ids to include from full corpus, all loaded if None 
    config_path (str) : path to the configuration file containing various parameters
    tokenzier (str) : name of the tokenizer model
    """
    def __init__(self,
        dataset: str,
        subset_ids:List,
        tokenizer: str = "bert-base-uncased",
        config_path: str = "config.ini",
        ):
        self.raw_data:List[Evidence] = []
        self.subset_ids = subset_ids
        self.config = configparser.ConfigParser()
        self.config.read(config_path)
        self.data_path = self.config["Data-Path"][dataset]  
        self.tokenizer_name = tokenizer      
        self._load_data()

        if(self.tokenizer_name):
            self.tokenized_path = f"{dataset}_{tokenizer}_tokenized"
            self.tokenizer_name = tokenizer     
            self.tokenizer = Tokenizer(self.tokenizer_name)            
            tokenized_data = self._load_tokenized_data()
            self.dataset = PassageDataset(tokenized_data['input_ids'].keys(),tokenized_data['input_ids'],tokenized_data['attention_mask'])
    
    def _load_data(self):
        if(".json" in self.data_path):
            with open(self.data_path,'r') as fp:
                db = json.load(fp) #format {"id":{"passage":"..","title":...}
                passages = {}
                titles = {}
                types = {}
                for idx in tqdm(db.keys(),total=len(db),desc="Loading passages"):
                    if "text" in db[idx].keys():
                        passages[idx] = db[idx]["text"]
                    else:
                        passages[idx] = db[idx]["passage"]
                    titles[idx] = db[idx]["title"] if db[idx]["title"] else ""
                    if("type" in db[idx].keys()):
                        types[idx] = db[idx]["type"]
        else:
            passages,titles = self.load_passage_db(self.data_path,copy.copy(self.subset_ids))
        subset_ids = self.subset_ids
        if not(subset_ids):
            subset_ids = list(passages.keys())
        for i in tqdm(range(len(subset_ids)),total = len(subset_ids),desc="Transforming passage dataset"):
            idx = str(subset_ids[i])
            if(idx in types.keys() and types[idx]==DataTypes.TABLE):
                rows = passages[idx].split(Separators.TABLE_ROW_SEP)
                columns = rows[0].split(Separators.TABLE_COL_SEP)
                table = [row.split(Separators.TABLE_COL_SEP) for row in rows[1:]]
                self.raw_data.append(TableEvidence(idx=idx,title=titles[idx],columns=columns,table=table))
            else:
                self.raw_data.append(Evidence(text=passages[str(subset_ids[i])],idx=subset_ids[i],title=titles[str(subset_ids[i])])) 

    
    def _load_tokenized_data(self):
        if self.tokenized_path and os.path.exists(self.tokenized_path):
            self.logger.info("Loading DPR data from {}".format(self.tokenized_path))
            with open(self.tokenized_path, "r") as fp:
                return json.load(fp)
        else:             
            input_data = [passage.title() + " " + "[SEP]" + " " + passage.text() for passage in self.raw_data]
            psg_ids = [passage.id() for passage in self.raw_data]
            tokenized_data = self.tokenizer.tokenize(input_data,
                            max_length=128,
                            pad_to_max_length=True)
            input_ids = {_id: _input_ids
                            for _id, _input_ids in zip(psg_ids, tokenized_data["input_ids"])}
            attention_mask = {_id: _attention_mask
                                for _id, _attention_mask in zip(psg_ids, tokenized_data["attention_mask"])}
            return {"input_ids": input_ids, "attention_mask": attention_mask}            

    
    def load_passage_db(self,data_path, subset=None):
        passages = {}
        titles = {}
        load_all = True if not(subset) else False
        s_len = len(subset) if not(load_all) else None

        with gzip.open(data_path, "rb") as f:
            _ = f.readline()
            offset = 0
            for line in tqdm(f):
                if load_all or offset in subset:
                    _id, passage, title = line.decode().strip().split("\t")
                    assert int(_id) - 1 == offset
                    passages[offset] = passage.lower()
                    titles[offset] = title.lower()
                    subset.remove(offset)
                    if(not(subset)):
                        break
                offset += 1
        if(not(load_all)):
            assert s_len == len(titles) == len(passages)
        print("Loaded {} passages".format(len(passages)))
        return passages,titles 
        

    






from typing import List


class Answer:
    """
    A base class to hold all details about the answer aspect in question answering.

    Attributes:
        _text (str): The text of the answer.
        _idx (int): The ID of the answer.
    """
    def __init__(self, text: str, idx=None):
        self._text = text
        self._idx = idx

    def text(self)->str:
        return self._text

    def id(self):
        return self._idx

    def flatten(self)->List[str]:
        """
        Flattends the answer structure if complex into a simple list of answer texts.
        """
        return [self._text]

    # def set_id(self, id):
    #     self._id = id
    #
    # def set_attention_mask(self, attention_mask):
    #     self.attention_mask = attention_mask


class AmbigNQAnswer:
    """
    The Ambiguous question answer case contains questions which can be divided into sub-clarified questions with corresponding answers.
    Additionally, each question can have multiple annotations. 
    The element answers[i][j][k] represents the answer given by the ith annotator for the jth clarified question and the kth answer.

    Attributes:
        answers  List[List[List[Answer]]]: The text of the answer given in structure mentioned above.
        _idx id of the answer
    """
    def __init__(self, answers: List[List[List[Answer]]],idx=None):
        self.answers = answers
        self._idx = idx
    
    def id(self):
        return self._idx

    def flatten(self)->List[str]:
        """
        Flattends the answer structure into a simple list of answer texts.
        """
        flattened_answers = []
        for annotation in self.answers:
            for query in annotation:
                for answer in query:
                    flattened_answers.append(answer.text())
        return flattened_answers

class TATQAAnswer(Answer):
    """
    The TATQAAnswer question answer case contains questions qhich contains a sequnce of answers for one question.
    The element answers[i] represents the ith answer for the question.

    Attributes:
        answers  List[Answer]: The text of the answer given in structure mentioned above.
        _idx id of the answer
    """

    def __init__(self, answers: List[Answer], idx=None):
        self.answers = answers
        self.idx = idx

    def text(self):
        return ",".join(self._text)

    def id(self):
        return self._idx



from configparser import ConfigParser

from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from dexter.retriever.dense.DenseFullSearch import DenseFullSearch



class ANCE(DenseFullSearch):
    #Wrapper for ANCE on top of Dense.
    # Will be different from parent when train loop implemented

    def __init__(self,config_path="config.ini",show_progress_bar=True,convert_to_tensor=True,batch_size=32) -> None:        
        self.config = ConfigParser()
        self.config.read(config_path)
        question_encoder = self.config["Retrieval"]["question-encoder"]
        context_encoder = self.config["Retrieval"]["context-encoder"]
        show_progress_bar = show_progress_bar
        convert_to_tensor = convert_to_tensor
        batch_size = batch_size
        dense_hyperparams = DenseHyperParams(query_encoder_path=question_encoder,document_encoder_path=context_encoder,show_progress_bar=show_progress_bar,convert_to_tensor=convert_to_tensor,batch_size=batch_size)        
        super().__init__(dense_hyperparams)


class Split:
    TEST = "test"
    TRAIN = "train"
    DEV = "dev"
    PRED = "predict"

class Dataset:
    AMBIGQA = "ambignq"
    WIKIMULTIHOPQA = "wikimultihopqa"
    FINQA = "finqa"
    TATQA = "tatqa"
    MUSIQUEQA="musiqueqa"
    OTTQA = "ottqa"
    StrategyQA = "strategyqa"

class Separators:
    TABLE_ROW_SEP = '\n'
    TABLE_COL_SEP = '|'

class DataTypes:
    TABLE = "table"
    TEXT = "text"



from elasticsearch import Elasticsearch
from elasticsearch.helpers import streaming_bulk
from typing import Dict, List, Tuple
import logging

tracer = logging.getLogger('elasticsearch') 
tracer.setLevel(logging.CRITICAL) # supressing INFO messages for elastic-search

class ElasticSearch(object):
    
    def __init__(self, es_credentials: Dict[str, object]):
        
        logging.info("Activating Elasticsearch....")
        logging.info("Elastic Search Credentials: %s", es_credentials)
        self.index_name = es_credentials["index_name"]
        self.check_index_name()

        # https://www.elastic.co/guide/en/elasticsearch/reference/current/analysis-lang-analyzer.html
        self.languages = ["arabic", "armenian", "basque", "bengali", "brazilian", "bulgarian", "catalan", 
                          "cjk", "czech", "danish", "dutch", "english","estonian","finnish","french", 
                          "galician", "german", "greek", "hindi", "hungarian", "indonesian", "irish", 
                          "italian", "latvian", "lithuanian", "norwegian", "persian", "portuguese", 
                          "romanian", "russian", "sorani", "spanish", "swedish", "turkish", "thai"]
        
        self.language = es_credentials["language"]
        self.check_language_supported()

        self.text_key = es_credentials["keys"]["body"]
        self.title_key = es_credentials["keys"]["title"]
        self.number_of_shards = es_credentials["number_of_shards"]
        
        self.es = Elasticsearch(
            [es_credentials["hostname"]], 
            timeout=es_credentials["timeout"], 
            retry_on_timeout=es_credentials["retry_on_timeout"], 
            maxsize=es_credentials["maxsize"],
            ca_certs=es_credentials["ca_certs"],
            http_auth=es_credentials["basic_auth"])

    def check_language_supported(self):
        """Check Language Supported in Elasticsearch
        """
        if self.language.lower() not in self.languages:
            raise ValueError("Invalid Language: {}, not supported by Elasticsearch. Languages Supported: \
            {}".format(self.language, self.languages))
    
    def check_index_name(self):
        """Check Elasticsearch Index Name"""
        # https://stackoverflow.com/questions/41585392/what-are-the-rules-for-index-names-in-elastic-search
        # Check 1: Must not contain the characters ===> #:\/*?"<>|,
        for char in '#:\/*?"<>|,':
            if char in self.index_name:
                raise ValueError('Invalid Elasticsearch Index, must not contain the characters ===> #:\/*?"<>|,')
        
        # Check 2: Must not start with characters ===> _-+
        if self.index_name.startswith(("_", "-", "+")):
            raise ValueError('Invalid Elasticsearch Index, must not start with characters ===> _ or - or +')
        
        # Check 3: must not be . or ..
        if self.index_name in [".", ".."]:
            raise ValueError('Invalid Elasticsearch Index, must not be . or ..')
        
        # Check 4: must be lowercase
        if not self.index_name.islower():
            raise ValueError('Invalid Elasticsearch Index, must be lowercase')
        
    
    def create_index(self):
        """Create Elasticsearch Index
        """
        logging.info("Creating fresh Elasticsearch-Index named - {}".format(self.index_name))
        
        try:
            if self.number_of_shards == "default":
                mapping = {
                    "mappings" : {
                        "properties" : {
                            self.title_key: {"type": "text", "analyzer": self.language},
                            self.text_key: {"type": "text", "analyzer": self.language}
                        }}}
            else:
                mapping = {
                    "settings": {
                        "number_of_shards": self.number_of_shards
                    },
                    "mappings" : {
                        "properties" : {
                            self.title_key: {"type": "text", "analyzer": self.language},
                            self.text_key: {"type": "text", "analyzer": self.language}
                        }}}
                
            self.es.indices.create(index=self.index_name, body=mapping, ignore=[400]) #400: IndexAlreadyExistsException
        except Exception as e:
            logging.error("Unable to create Index in Elastic Search. Reason: {}".format(e))
    
    def delete_index(self):
        """Delete Elasticsearch Index"""
        
        logging.info("Deleting previous Elasticsearch-Index named - {}".format(self.index_name))
        try:
            self.es.indices.delete(index=self.index_name, ignore=[400, 404]) # 404: IndexDoesntExistException
        except Exception as e:
            logging.error("Unable to create Index in Elastic Search. Reason: {}".format(e))
    
    def bulk_add_to_index(self, generate_actions, progress):
        """Bulk indexing to elastic search using generator actions

        Args:
            generate_actions (generator function): generator function must be provided 
            progress (tqdm.tqdm): tqdm progress_bar
        """
        for ok, action in streaming_bulk(
        client=self.es, index=self.index_name, actions=generate_actions,
        ):
            progress.update(1)
        progress.reset()
        progress.close()
    
    def lexical_search(self, text: str, top_hits: int, ids: List[str] = None, skip: int = 0) -> Dict[str, object]:
        """[summary]

        Args:
            text (str): query text
            top_hits (int): top k hits to retrieved
            ids (List[str], optional): Filter results for only specific ids. Defaults to None.

        Returns:
            Dict[str, object]: Hit results
        """
        req_body = {"query" : {"multi_match": {
                "query": text, 
                "type": "best_fields",
                "fields": [self.text_key, self.title_key],
                "tie_breaker": 0.5
                }}}
        
        if ids: req_body = {"query": {"bool": {
                    "must": req_body["query"],
                    "filter":  {"ids": {"values": ids}}
                }}}

        res = self.es.search(
            search_type="dfs_query_then_fetch",
            index = self.index_name, 
            body = req_body, 
            size = skip + top_hits
        )
        
        hits = []
        
        for hit in res["hits"]["hits"][skip:]:
            hits.append((hit["_id"], hit['_score']))
        
        return self.hit_template(es_res=res, hits=hits)
    
    
    def lexical_multisearch(self, texts: List[str], top_hits: int, skip: int = 0) -> Dict[str, object]:
        """Multiple Query search in Elasticsearch

        Args:
            texts (List[str]): Multiple query texts
            top_hits (int): top k hits to be retrieved
            skip (int, optional): top hits to be skipped. Defaults to 0.

        Returns:
            Dict[str, object]: Hit results
        """
        request = []
        
        assert skip + top_hits <= 10000, "Elastic-Search Window too large, Max-Size = 10000"
        
        for text in texts:
            req_head = {"index" : self.index_name, "search_type": "dfs_query_then_fetch"}
            req_body = {
                "_source": False, # No need to return source objects
                "query": {
                    "multi_match": { 
                        "query": text, # matching query with both text and title fields
                        "type": "best_fields",
                        "fields": [self.title_key, self.text_key],
                        "tie_breaker": 0.5
                        }
                    },
                "size": skip + top_hits, # The same paragraph will occur in results
                }
                
            request.extend([req_head, req_body])
        
        res = self.es.msearch(body = request)

        result = []
        for resp in res["responses"]:
            #print("resp",resp)
            responses = resp["hits"]["hits"][skip:]
            
            hits = []
            for hit in responses:
                hits.append((hit["_id"], hit['_score']))

            result.append(self.hit_template(es_res=resp, hits=hits))
        return result
    
    
    def generate_actions(self, dictionary: Dict[str, Dict[str, str]], update: bool = False):
        """Iterator function for efficient addition to Elasticsearch
        Ref: https://stackoverflow.com/questions/35182403/bulk-update-with-pythons-elasticsearch
        """
        for _id, value in dictionary.items():
            if not update:
                doc = {
                    "_id": str(_id),
                    "_op_type": "index",

                    "refresh": "wait_for",
                    self.text_key: value[self.text_key],
                    self.title_key: value[self.title_key],
                }
            else:
                doc = {
                    "_id": str(_id),
                    "_op_type": "update",
                    "refresh": "wait_for",

                    "doc": {
                        self.text_key: value[self.text_key],
                        self.title_key: value[self.title_key],
                        }
                }
                
            yield doc
        
    def hit_template(self, es_res: Dict[str, object], hits: List[Tuple[str, float]]) -> Dict[str, object]:
        """Hit output results template

        Args:
            es_res (Dict[str, object]): Elasticsearch response
            hits (List[Tuple[str, float]]): Hits from Elasticsearch

        Returns:
            Dict[str, object]: Hit results
        """
        result = {
            'meta': {
                'total': es_res['hits']['total'],
                'took': es_res['took'],
                'num_hits': len(hits)
            },
            'hits': hits,
        }
        return result


from transformers import AutoModelForSeq2SeqLM, AutoTokenizer


class FlanT5Engine:
    def __init__(self, data,model_name="google/flan-t5-xl", temperature=None, top_n=None):
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name)
        self.model.cuda()
        self.temperature = temperature
        self.data = data

    def get_flant5_completion(self, prompt: str):
        input = self.tokenizer(prompt, return_tensors="pt",
                               max_length=1096).to("cuda")
        output = self.model.generate(**input, max_length=1000)
        decoded_outputs = self.tokenizer.batch_decode(output,
                                                      skip_special_tokens=True, max_length=1000)[0]
        return decoded_outputs



import torch
import transformers
from transformers import AutoModelForCausalLM, AutoTokenizer


class LlamaEngine:
    def __init__(self, data, model_name="meta-llama/Llama-2-7b-hf", temperature=None, top_n=None):
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForCausalLM.from_pretrained(model_name)
        self.temperature = temperature
        self.data = data
        self.top_n = top_n

    def get_llama_completion(self, prompt: str):

        pipeline = transformers.pipeline(
            "text-generation",
            model=self.model,
            tokenizer=self.tokenizer,
            torch_dtype=torch.bfloat16,
            trust_remote_code=True,
            device_map="auto",
            max_length=1000,
            do_sample=True,
            top_k=self.top_n,
            num_return_sequences=1,
            eos_token_id=self.tokenizer.eos_token_id
        )
        output = pipeline(prompt)
        return output[0]["generated_text"]



"""
 FAISS-based index components for dense retriver
"""

import logging

from dexter.retriever.dense.indexer.annoyIndexer import AnnoySearch
from dexter.retriever.dense.indexer.faissIndexer import FaissSearch


logger = logging.getLogger()

from dexter.config import AnnConstants

class AnnSearch:
    @staticmethod
    def get_ann_instance(ann_class, data, emb_dim,num_trees=None, metric=None):
        if ann_class.lower() == AnnConstants.FAISS_SEARCH.lower():
            return FaissSearch(data,emb_dim)

        elif ann_class.lower() == AnnConstants.ANNOY_SEARCH.lower():
            return AnnoySearch(data,num_trees,emb_dim,metric)

        else:
            classname = ann_class
            if classname not in globals():
                raise ValueError("No implementation found for the custom data generator class specified: {}".format(classname))
            return globals()[classname]()



import pytrec_eval
from typing import List, Dict, Tuple
from dexter.utils.metrics.retrieval.accuracy import top_k_accuracy
import logging

logger = logging.getLogger(__name__)

class RetrievalMetrics:
    """retrieval metrics ndcg, mrr etc.
    """    
    def __init__(self,  k_values: List[int] = [1,3,5,10,100]):
        self.k_values = k_values
        self.top_k = max(k_values)

    def evaluate_retrieval(self,qrels: Dict[str, Dict[str, int]], 
                 results: Dict[str, Dict[str, float]]):
        """_summary_

        Args:
            qrels (Dict[str, Dict[str, int]]): query doc relevance mapping for evaluation of results
            results (Dict[str, Dict[str, float]]): _description_

        Returns:
            Tuple[Dict[str, float], Dict[str, float], Dict[str, float], Dict[str, float]]: _description_
        """        


        ndcg = {}
        _map = {}
        recall = {}
        precision = {}
        accuracy = {}
        
        for k in self.k_values:
            ndcg[f"NDCG@{k}"] = 0.0
            _map[f"MAP@{k}"] = 0.0
            recall[f"Recall@{k}"] = 0.0
            precision[f"P@{k}"] = 0.0
            accuracy[f"acc@{k}"] = 0.0
        
        map_string =  set(["map_cut."+str(k) for k in self.k_values])
        ndcg_string = set(["ndcg_cut."+str(k) for k in self.k_values])
        recall_string =  set(["recall."+str(k) for k in self.k_values])
        precision_string =  set(["P."+str(k) for k in self.k_values])
        evaluator = pytrec_eval.RelevanceEvaluator(qrels, map_string|ndcg_string|recall_string|precision_string)
        scores = evaluator.evaluate(results)
        
        for query_id in scores.keys():
            for k in self.k_values:
                ndcg[f"NDCG@{k}"] += scores[query_id]["ndcg_cut_" + str(k)]
                _map[f"MAP@{k}"] += scores[query_id]["map_cut_" + str(k)]
                recall[f"Recall@{k}"] += scores[query_id]["recall_" + str(k)]
                precision[f"P@{k}"] += scores[query_id]["P_"+ str(k)]
        
        for k in self.k_values:
            ndcg[f"NDCG@{k}"] = round(ndcg[f"NDCG@{k}"]/len(scores), 5)
            _map[f"MAP@{k}"] = round(_map[f"MAP@{k}"]/len(scores), 5)
            recall[f"Recall@{k}"] = round(recall[f"Recall@{k}"]/len(scores), 5)
            precision[f"P@{k}"] = round(precision[f"P@{k}"]/len(scores), 5)
        
        for eval in [ndcg, _map, recall, precision]:
            logger.info("\n")
            for k in eval.keys():
                logger.info("{}: {:.4f}".format(k, eval[k]))
        accuracy[f"acc@{k}"] = top_k_accuracy(qrels, results, self.k_values)

        return ndcg, _map, recall, precision


from torch.utils.data import Dataset



class QADataset(Dataset):
    """
    QA dataset for tokenized questions and corresponding answers to train answering models
    """
    def __init__(self, enc_ids, enc_mask, dec_ids, dec_mask, is_training):
        assert len(enc_ids) == len(enc_mask)
        assert len(dec_ids) == len(dec_mask)
        self.enc_ids = enc_ids
        self.enc_mask = enc_mask
        self.dec_ids = dec_ids
        self.dec_mask = dec_mask
        self.is_training = is_training

    def __len__(self):
        return len(self.enc_ids)

    def __getitem__(self, idx):
        if self.is_training:
            return self.enc_ids[idx], self.enc_mask[idx], self.dec_ids[idx], self.dec_mask[idx]
        else:
            return self.enc_ids[idx], self.enc_mask[idx]


class DprDataset(Dataset):
    """ Dataset to hold tokenized Question and Related Context

    Args:
        query_ids : input ids of all tokenized questions
        query_mask : attention mask for tokenized question
        context_ids : input ids of tokenized context
        context_mask : attention mask for tokenized context
    """
    def __init__(self, query_ids, query_mask, context_ids, context_mask):
        assert len(query_ids) == len(query_mask)
        assert len(context_ids) == len(context_mask)
        self.query_ids = query_ids
        self.query_mask = query_mask
        self.context_ids = context_ids
        self.context_mask = context_mask

    def __len__(self):
        return len(self.query_ids)

    def __getitem__(self, idx):
        return self.query_ids[idx], self.query_mask[idx], self.context_ids[idx], self.context_mask[idx]

class PassageDataset(Dataset):
    """ Dataset to load and process an corpus containing passages that can serve as contexts for question answering.

    Args:
        passage_ids : ids of passages in the corpus for identification
        input_ids : input ids of context post tokenization
        attention_mask : attention mask for context post tokenization
        """
    def __init__(self,passage_ids,input_ids,attention_mask):
        assert len(input_ids)==len(passage_ids)==len(attention_mask)
        self.input_ids = input_ids
        self.passage_ids = passage_ids
        self.attention_mask = attention_mask

    def __len__(self):
        return len(self.passage_ids)

    def __getitem__(self, idx):
        assert idx in self.passage_ids
        return self.input_ids[idx], self.attention_mask[idx]

    def get_by_id(self,idx):
        return self.__getitem__(idx)



import logging
from typing import List, Dict, Tuple


'''This file contains function to calculate accuracy '''
def top_k_accuracy(
        qrels: Dict[str, Dict[str, int]], 
        results: Dict[str, Dict[str, float]], 
        k_values: List[int]) -> Tuple[Dict[str, float]]:
    
    top_k_acc = {}
    
    for k in k_values:
        top_k_acc[f"Accuracy@{k}"] = 0.0
    
    k_max, top_hits = max(k_values), {}
    logging.info("\n")
    
    for query_id, doc_scores in results.items():
        top_hits[query_id] = [item[0] for item in sorted(doc_scores.items(), key=lambda item: item[1], reverse=True)[0:k_max]]
    
    for query_id in top_hits:
        query_relevant_docs = set([doc_id for doc_id in qrels[query_id] if qrels[query_id][doc_id] > 0])
        for k in k_values:
            for relevant_doc_id in query_relevant_docs:
                if relevant_doc_id in top_hits[query_id][0:k]:
                    top_k_acc[f"Accuracy@{k}"] += 1.0
                    break

    for k in k_values:
        top_k_acc[f"Accuracy@{k}"] = round(top_k_acc[f"Accuracy@{k}"]/len(qrels), 5)
        logging.info("Accuracy@{}: {:.4f}".format(k, top_k_acc[f"Accuracy@{k}"]))

    return top_k_acc


import os
import faiss
import logging
from typing import List, Dict
from torch import Tensor

logger = logging.getLogger(__name__)
dir_path = os.path.dirname(os.path.realpath(__file__))


class FaissSearch:
    def __init__(self,data: List[str]=None,emb_dim=None):
        self.emb_dim=emb_dim
        self.data = data
        self.ann = faiss.IndexFlatIP(self.emb_dim)


    def get_top_n_neighbours(self,query_vector: Tensor, top_k: int)->Dict:
        print("query_vector",query_vector.shape)
        query_vector = query_vector
        distances, indices = self.ann.search(query_vector,top_k)
        assert indices.shape == distances.shape == (query_vector.shape[0],top_k)
        indices = indices.tolist()
        #passages = [self.data[idx] for passage_ids in indices for idx in passage_ids]
        return {"ids": indices, "distances": distances}

    def load_index_if_available(self)->None:
        if os.path.exists("indices/faiss/index_faiss"):
            self.ann = faiss.read_index("indices/faiss/index_faiss")
            return True
        else:
            return False
    def create_index(self, passage_vectors):
        self.ann.add(passage_vectors)
        if not os.path.exists("indices/faiss"):
            os.makedirs("indices/faiss")
        faiss.write_index(self.ann, "indices/faiss/index_faiss")




from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from dexter.retriever.dense.HfRetriever import HfRetriever

class Contriever(HfRetriever):
    #Wrapper class for future extensions
    def __init__(self,config=DenseHyperParams) -> None:
        super().__init__(config)
        self.config = config


from sentence_transformers.cross_encoder import CrossEncoder as CrossEnc
from typing import List, Tuple, Dict

class CrossEncoder:
    def __init__(self, model_name: str, **kwargs):
        self.model = CrossEnc(model_name, **kwargs)
    
    def predict(self, sentences: List[Tuple[str,str]], 
                batch_size: int = 32, 
                show_progress_bar: bool = True) -> List[float]:
        return self.model.predict(
            sentences=sentences, 
            batch_size=batch_size, 
            show_progress_bar=show_progress_bar)


from configparser import ConfigParser
import heapq
from typing import Dict, List, Union, Tuple, Any
import numpy as np
import os
import heapq
import joblib
from torch import Tensor
import torch
from tqdm import tqdm
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question

from transformers import AutoTokenizer, AutoModel
import logging
from dexter.data.datastructures.hyperparameters.dpr import DenseHyperParams
from dexter.retriever.BaseRetriever import BaseRetriver
from dexter.utils.metrics.SimilarityMatch import SimilarityMetric

class HfRetriever(BaseRetriver):

    def __init__(self,config=DenseHyperParams) -> None:
        super().__init__()
        self.config = config
        #print("self.config.query_encoder_path",self.config.query_encoder_path)
        self.question_tokenizer = AutoTokenizer.from_pretrained(self.config.query_encoder_path)
        self.context_tokenizer = AutoTokenizer.from_pretrained(self.config.document_encoder_path)
        self.question_encoder = AutoModel.from_pretrained(self.config.query_encoder_path)
        self.context_encoder = AutoModel.from_pretrained(self.config.document_encoder_path)
        self.question_encoder.cuda()
        self.context_encoder.cuda()
        self.batch_size = self.config.batch_size
        self.sep="[SEP]"
        self.logger = logging.getLogger(__name__)

    def mean_pooling(self, token_embeddings, mask):
        token_embeddings = token_embeddings.masked_fill(~mask[..., None].bool(), 0.)
        sentence_embeddings = token_embeddings.sum(dim=1) / mask.sum(dim=1)[..., None]
        return sentence_embeddings


    def encode_queries(self, 
                       queries: List[Question], 
                       batch_size: int = 16,
                         **kwargs) -> Union[List[Tensor], np.ndarray, Tensor]:
        with torch.no_grad():
            tokenized_questions = self.question_tokenizer([query.text() for query in queries], padding=True, truncation=True, return_tensors='pt').to("cuda")
            token_emb =  self.question_encoder(**tokenized_questions)
        print("token_emb",token_emb[0].shape)
        sentence_emb = self.mean_pooling(token_emb[0],tokenized_questions["attention_mask"])
        print("sentence_emb",sentence_emb.shape)
        return sentence_emb
    
    def encode_corpus(self, 
                      corpus: List[Evidence], 
                      **kwargs
                      ) -> Union[List[Tensor], np.ndarray, Tensor]:
        contexts = []
        for evidence in corpus:
            context = ""
            if evidence.title():
                context = (evidence.title() + self.sep + evidence.text()).strip()
            else:
                context = evidence.text().strip()
            contexts.append(context)
        context_embeddings = []
        index = 0
        pbar = tqdm(total = len(contexts))
        print("Starting encoding of contexts....")
        with torch.no_grad():
            while index < len(contexts):
                samples = contexts[index:index+self.batch_size]
                tokenized_contexts = self.context_tokenizer(samples, padding=True, truncation=True, return_tensors='pt').to("cuda")
                token_emb =  self.context_encoder(**tokenized_contexts)
                sentence_emb = self.mean_pooling(token_emb[0],tokenized_contexts["attention_mask"])
                context_embeddings.append(sentence_emb)
                index += self.batch_size
                pbar.update(self.batch_size)
        pbar.close()
        context_embeddings = torch.cat(context_embeddings,dim=0)
        print("context_embeddings",context_embeddings.shape)
        return context_embeddings

    def load_index_if_available(self)->Tuple[Any,bool]:
        if os.path.exists("indices/corpus/index"):
            corpus_embeddings = joblib.load("indices/corpus/index")
            index_present=True
        else:
            index_present = False
            corpus_embeddings=None
        return corpus_embeddings, index_present

    def retrieve_in_chunks(self,
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True,
                chunksize: int =200000,
                show_progress_bar =True,
                convert_to_tensor = True,
                  **kwargs  ):
        corpus_ids = [doc.id() for doc in corpus]
        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size)  
        query_ids = [query.id() for query in queries]
        result_heaps = {qid: [] for qid in query_ids}  # Keep only the top-k docs for each query
        self.results = {qid: {} for qid in query_ids}
        batches = range(0, len(corpus), chunksize)
        for batch_num, corpus_start_idx in enumerate(batches):
            self.logger.info("Encoding Batch {}/{}...".format(batch_num+1, len(batches)))
            corpus_end_idx = min(corpus_start_idx + chunksize, len(corpus))

            # Encode chunk of corpus    
            sub_corpus_embeddings = self.encode_corpus(
                corpus[corpus_start_idx:corpus_end_idx]
                )

            # Compute similarites using either cosine-similarity or dot product
            cos_scores = score_function.evaluate(query_embeddings, sub_corpus_embeddings)
            cos_scores[torch.isnan(cos_scores)] = -1

            # Get top-k values
            cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
            cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
            cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
            
            for query_itr in range(len(query_embeddings)):
                query_id = query_ids[query_itr]                  
                for sub_corpus_id, score in zip(cos_scores_top_k_idx[query_itr], cos_scores_top_k_values[query_itr]):
                    corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                    if corpus_id != query_id:
                        if len(result_heaps[query_id]) < top_k:
                            # Push item on the heap
                            heapq.heappush(result_heaps[query_id], (score, corpus_id))
                        else:
                            # If item is larger than the smallest in the heap, push it on the heap then pop the smallest element
                            heapq.heappushpop(result_heaps[query_id], (score, corpus_id))

        for qid in result_heaps:
            for score, corpus_id in result_heaps[qid]:
                self.results[qid][corpus_id] = score
        return self.results

    def retrieve(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               score_function: SimilarityMetric,
               return_sorted: bool = True, 
               chunk: bool = False,
               chunksize = None,
               **kwargs) -> Dict[str, Dict[str, float]]:

            
        self.logger.info("Encoding Queries...")

        query_embeddings = self.encode_queries(queries, batch_size=self.batch_size)
          
        if chunk:
            results = self.retrieve_in_chunks(corpus, 
                                              queries,top_k=top_k,
                                              score_function=score_function,return_sorted=return_sorted,
                                              chunksize=chunksize)
            return results
   
        self.logger.info("Encoding Corpus in batches... Warning: This might take a while!")
        #self.logger.info("Scoring Function: {} ({})".format(self.score_function_desc[score_function], score_function))
        embeddings, index_present = self.load_index_if_available()

        #TODO:Comment below for index usage
        index_present = False
        if index_present:
            corpus_embeddings = embeddings
        else:
            corpus_embeddings = self.encode_corpus(corpus)
            #joblib.dump(corpus_embeddings,"indices/corpus/index")

        # Compute similarites using either cosine-similarity or dot product
        cos_scores = score_function.evaluate(query_embeddings,corpus_embeddings)
        # Get top-k values
        cos_scores_top_k_values, cos_scores_top_k_idx = torch.topk(cos_scores, min(top_k+1, len(cos_scores[1])), dim=1, largest=True, sorted=return_sorted)
        cos_scores_top_k_values = cos_scores_top_k_values.cpu().tolist()
        cos_scores_top_k_idx = cos_scores_top_k_idx.cpu().tolist()
        response = {}
        for idx, q in enumerate(queries):
            response[q.id()] = {}
            for index, id in enumerate(cos_scores_top_k_idx[idx]):
                response[q.id()][corpus[id].id()] = float(cos_scores_top_k_values[idx][index])
        return response
                


import json
import os
import tqdm
from dexter.config.constants import Split
from dexter.data.datastructures.answer import Answer
from dexter.data.datastructures.dataset import DprDataset
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
from dexter.data.datastructures.sample import Sample
from dexter.data.loaders.BaseDataLoader import GenericDataLoader


class WikiMultihopQADataLoader(GenericDataLoader):
    '''Data loader class to load Datset from raw WikiMultihopQA dataset.
    WikiMultihopQA dataset consists of a questions each of which have a single answer and various evidences.
    
    Arguments:
    dataset (str): string containing the dataset alias
    tokenzier (str) : name of the tokenizer model. Set tokenizer as None, if only samples to be loaded but not tokenized and stored. This can help save time if only the raw dataset is needed.
    config_path (str) : path to the configuration file containing various parameters
    split (Split) : Split of the dataset to be loaded
    batch_size (int) : batch size to process the dataset.
    corpus Dict[str,Evidence]: corpus containing all needed passages.    
    '''
    def __init__(self, dataset: str, tokenizer="bert-base-uncased", config_path='test_config.ini', split=Split.TRAIN,
                 batch_size=None, corpus=None):
        self.corpus = corpus
        self.titles = [self.corpus[idx].title().split(" - ")[0] for idx,_ in enumerate(self.corpus)]
        print(self.titles[100],self.corpus[100].title())
        super().__init__(dataset, tokenizer, config_path, split, batch_size)


    def load_raw_dataset(self, split=Split.TRAIN):
        dataset = self.load_json(split)
        print(len(dataset))
        for  query_index, data in enumerate(tqdm.tqdm(dataset[:1200])):
            if len(data["context"]) == 0:
                data["context"] = ['some random title', ['some random stuff']]
            for evidence_set in data["context"]:
                title = evidence_set[0]
                #for evidence in evidence_set[1]:
                evidence = " ".join(evidence_set[1])
                #print(list(self.titles).index(title.split(" - ")[0]))
                self.raw_data.append(
                    Sample(query_index, Question(data["question"],idx=data["_id"]), Answer(data["answer"]),
                            Evidence(text=evidence, 
                                    idx=list(self.titles).index(title.split(" - ")[0]),title=title)
                ))

    def load_tokenized(self):
        if self.tokenized_path and os.path.exists(self.tokenized_path):
            self.logger.info("Loading DPR data from {}".format(self.tokenized_path))
            with open(self.tokenized_path, "r") as f:
                ip_ids, ip_attention, evidence_ids, evidence_attention = json.load(f)
        else:
            ip_ids, ip_attention = self.tokenize_questions()
            evidence_ids, evidence_attention = self.tokenize_evidences()
        return DprDataset(ip_ids, ip_attention, evidence_ids, evidence_attention)


from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

optional_packages = {
    "tf" : ['tensorflow>=2.2.0', 'tensorflow-text', 'tensorflow-hub']
}

setup(
    name="dexter-cqa",
    version="1.0.3",
    author="Venktesh V, Deepali Prabhu",
    author_email="venkyviswa12@gmail.com",
    description="A Benchmark for Complex Heterogeneous Question answering",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    url="https://github.com/VenkteshV/BCQA",
    download_url="",
    packages=find_packages(),
    python_requires='>=3.8',
    install_requires=[
        'sentence-transformers',
        'pytrec_eval',
        'faiss_cpu',
        'elasticsearch==7.9.1',
        'data',
        'toml',
        'zope.interface',
        'transformers==4.30.0',
        'protobuf',
        'openai',
        "annoy",
        'pytrec_eval'
    ],
    extras_require = optional_packages,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    keywords="Information Retrieval Transformer Networks Complex Question Answering BERT PyTorch Question Answering IR NLP deep learning"
)


import torch
import transformers
from transformers import AutoModelForCausalLM, AutoTokenizer


class MistralEngine:

    def __init__(self, data, model_name="mistralai/Mistral-7B-Instruct-v0.1", temperature=0.3, top_n=1, max_new_tokens=256):
        self.model_name = model_name
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.temperature = temperature
        self.data = data
        self.top_n = top_n
        self.max_new_tokens=max_new_tokens
        self.pipeline = transformers.pipeline(
            "text-generation",
            model=self.model_name,
            torch_dtype=torch.float16,
            device_map="auto",
        )

    def get_mistral_completion(self, system_prompt: str, user_prompt: str):
        messages = [
            {
                "role": "user",
                "content": system_prompt,
            },
              {
                "role": "assistant",
                "content": "Yes I will reason and generate the answer",
            },
            {"role": "user", "content": user_prompt
},
        ]
        prompt = self.pipeline.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        outputs = self.pipeline(prompt,
            max_new_tokens=self.max_new_tokens,
            do_sample=True,
            num_return_sequences=1,
            temperature=self.temperature,
            top_k=10,
            top_p=0.95)
        return outputs[0]["generated_text"]



"""
 LLM orchestrator factory that decides type of LLm engine to use
"""

import logging


from dexter.llms.flant5_engine import FlanT5Engine
from dexter.llms.llama_engine import LlamaEngine
from dexter.llms.mistral_engine import MistralEngine
from dexter.llms.openai_engine import OpenAIEngine

logger = logging.getLogger()
from dexter.config import LLmConstants

class LLMEngineOrchestrator:
    @staticmethod
    def get_llm_engine(llm_class, data,model_name=None, temperature=0.3,top_n = 1, top_p = None):
        if llm_class.lower() == LLmConstants.FLANT5.lower():
            return FlanT5Engine(data,model_name,temperature,top_n)

        elif llm_class.lower() == LLmConstants.OPENAI.lower():
            return OpenAIEngine(data,model_name,temperature,top_n)
        elif llm_class.lower() == LLmConstants.LLAMA.lower():
            return LlamaEngine(data, model_name, temperature, top_n)
        elif llm_class.lower() == LLmConstants.MISTRAL.lower():
            return MistralEngine(data,model_name,temperature,top_n)
        else:
            classname = llm_class
            if classname not in globals():
                raise ValueError("No implementation found for the custom class specified: {}".format(classname))
            return globals()[classname]()



class Question:
    """
    A base class to hold all details about the question aspect in question answering.

    Attributes:
        _text (str): The text of the question.
        _idx (int): The ID of the question.
    """

    def __init__(self, text:str, idx=None):
        self._text = text
        self._idx = idx
        self.attention_mask = None

    def text(self):
        return self._text

    def id(self):
        return self._idx

    def set_id(self, idx):
        self._id = idx

    def set_attention_mask(self, attention_mask):
        self.attention_mask = attention_mask



from dexter.config.constants import Dataset, Split
from dexter.data.loaders.AmbigQADataLoader import AmbigQADataLoader
from dexter.data.loaders.FinQADataLoader import FinQADataLoader
from dexter.data.loaders.MusiqueQaDataLoader import MusiqueQADataLoader

from dexter.data.loaders.OTTQADataLoader import OTTQADataLoader
from dexter.data.loaders.TATQADataLoader import TATQADataLoader
from dexter.data.loaders.WikiMultihopQADataLoader import WikiMultihopQADataLoader
from dexter.data.loaders.StrategyQADataLoader import StrategyQADataLoader



class DataLoaderFactory:
    ''' Data Loader factory to map dataset alias to corresponding Data loader class'''


    def create_dataloader(
        self,
        dataloader_name: str,
        tokenizer="bert-base-uncased",
        config_path="test_config.ini",
        split=Split.TRAIN,
        batch_size=None,
        corpus=None
    ):
        if Dataset.AMBIGQA in dataloader_name:
            loader = AmbigQADataLoader
        elif Dataset.FINQA in dataloader_name:
            loader = FinQADataLoader
        elif Dataset.WIKIMULTIHOPQA in dataloader_name:
            loader = WikiMultihopQADataLoader
        elif Dataset.MUSIQUEQA in dataloader_name:
            loader = MusiqueQADataLoader
        elif Dataset.TATQA in dataloader_name:
            loader = TATQADataLoader  
        elif Dataset.OTTQA in dataloader_name:
            loader = OTTQADataLoader
        elif Dataset.StrategyQA in dataloader_name:
            loader = StrategyQADataLoader
        else:
            raise NotImplemented(f"{dataloader_name} not implemented yet.")
        return loader(dataset=dataloader_name, config_path=config_path,
                      split=split,batch_size=batch_size,
                      tokenizer=tokenizer,
                      corpus=corpus)


'''This file contains the base hyperparameters for training downstream models.
Extend from this class to build custom hyperparameter class'''

import random
import numpy as np
import torch

class BaseHyperParameters:
    def __init__(self):
        self.adam_epsilon = 1e-8
        self.batch_size=16
        self.warmup =0.2 
    def set_seed(self, seed:int=None):
        if seed is None:
            seed = self.seed
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        if self.n_gpu > 0:
            torch.cuda.manual_seed_all(seed)


import re
from typing import Union, List
from math import isclose
class FinQAMatch:
    def _clean_num(self, text:str):
        text= str(text).replace("$","")

        text= str(text).replace("%","")
        text= str(text).replace(",","")

        return text

    def extract_num_from_str(self, answer: str):
        s = self._clean_num(answer)
        r_num = r"([+-]?\d+(\.\d+)?)|([+-]?\.\d+)"
        groups = re.findall(r_num, s)
        if len(groups) == 0:
            return None
        num = groups[-1][0]
        if num == '':
            return None
        if '.' in num:
            return float(num)
        return int(num)
    def finqa_equal(self,prediction: Union[bool, float, str],    
                reference: Union[float, str],
                include_percentage: bool = False,
                is_close: float = False) -> bool:
        if prediction is None:
            return False
        elif type(prediction) == bool:
            # bool questions
            if prediction:
                return reference == 'yes'
            else:
                return reference == 'no'

        else:
            # number questions
            gt_result = [reference]
            for item in gt_result:
                try:
                    #print(round(float(prediction), 1),round(float(item), 1))
                    if isclose(abs(round(float(item),3)), abs(round(float(prediction),3)), rel_tol=0.02):
                        return True
                    if round(abs(float(prediction)), 1) == round(abs(float(item)), 1):
                        return True
                except Exception as e:
                    print(e)
                    continue
            return False



import os
from annoy import AnnoyIndex
from torch import Tensor
import torch

from typing import List,Dict
import logging

logger = logging.getLogger(__name__)
dir_path = os.path.dirname(os.path.realpath(__file__))



class AnnoySearch:
    def __init__(self,data: List[str]=None, num_trees=None,emb_dim=None, metric="angular"):
        self.num_trees=num_trees
        self.emb_dim=emb_dim
        self.metric = metric
        self.data = data
        self.ann = AnnoyIndex(self.emb_dim, metric=self.metric)


    def get_top_n_neighbours(self,query_vector: Tensor, top_k: int, return_distances: bool = True)->Dict:
        matches = []
        all_distances = []
        for vec in query_vector:
            top_matches = self.ann.get_nns_by_vector(vec,top_k,include_distances=return_distances)
            indices = [idx for idx in top_matches[0]]
            distances = [dist for dist in top_matches[1]]
            matches.append(indices)
            all_distances.append(distances)


        return {"ids": indices, "distances": distances}

    def load_index_if_available(self)->None:
        if os.path.exists("indices/annoy/index.annoy"):
            self.ann.load("indices/annoy/index.annoy")
            return True
        else:
            return False

    def create_index(self, passage_vectors):
        for index, embed in enumerate(passage_vectors):
            self.ann.add_item(index, embed)
        self.ann.build(self.num_trees)
        if not os.path.exists("indices/annoy"):
            os.makedirs("indices/annoy")
        self.ann.save("indices/annoy/index.annoy")

if __name__=="__main__":
    annoy_instance = AnnoySearch()
    annoy_instance.get_index(torch.zeros(5,3))




from dexter.retriever.dense.ColBERT.colbert.infra.run import Run
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question
import logging
import tqdm
import heapq
from typing import List,Dict

from dexter.retriever.dense.ColBERT.build.lib.colbert.indexer import Indexer
from dexter.retriever.dense.ColBERT.colbert.infra.config.config import ColBERTConfig, RunConfig
from dexter.retriever.dense.ColBERT.colbert.searcher import Searcher



class TCTColBERT():
    #Wrapper class for future extensions
    def __init__(self,config=ColBERTConfig,
                 checkpoint=None) -> None:
        self.config = config
        self.indexer = Indexer(config=self.config,
                               checkpoint=checkpoint)
        self.sep="[SEP]"
        self.logger = logging.getLogger(__name__)

    def retrieve_in_chunks(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               return_sorted: bool = True,
                chunksize: int =200000, 
               **kwargs) -> Dict[str, Dict[str, float]]:
            print("retrieve in chunks")
            corpus_ids = [doc.id() for doc in corpus]
            corpus_texts = [(evidence.title() + self.sep + evidence.text()).strip() for evidence in corpus]
            queries_text = [question.text() for question in queries]
            query_ids = [query.id() for query in queries]
            result_heaps = {qid: [] for qid in query_ids}  # Keep only the top-k docs for each query
            self.results = {qid: {} for qid in query_ids}
            with Run().context(RunConfig(nranks=2, experiment="default")):
                batches = range(0, len(corpus), chunksize)
                print(batches)
                for batch_num, corpus_start_idx in enumerate(batches):
                    corpus_end_idx = min(corpus_start_idx + chunksize, len(corpus))

                    self.indexer.index(name="ambigqa", 
                    collection=corpus_texts[corpus_start_idx:corpus_end_idx], 
                    overwrite=True)

                    self.indexer.get_index()
                    searcher = Searcher(index="ambigqa", 
                    collection=corpus_texts[corpus_start_idx:corpus_end_idx])
                    for query_itr in range(len(queries)):
                        query_id = query_ids[query_itr] 
                        results = searcher.search(queries[query_itr].text(), k=top_k)                 
                        for sub_corpus_id, corpus_rank, score in zip(*results):
                            print("sub_corpus_id*********",len(corpus_ids),corpus_start_idx,sub_corpus_id)
                            corpus_id = corpus_ids[corpus_start_idx+sub_corpus_id]
                            if corpus_id != query_id:
                                if len(result_heaps[query_id]) < top_k:
                                    # Push item on the heap
                                    heapq.heappush(result_heaps[query_id], (score, corpus_id))
                                else:
                                    # If item is larger than the smallest in the heap, push it on the heap then pop the smallest element
                                    heapq.heappushpop(result_heaps[query_id], (score, corpus_id))

            for qid in result_heaps:
                for score, corpus_id in result_heaps[qid]:
                    self.results[qid][corpus_id] = score
                #results = searcher.search_all(queries_text,k=100)
            return self.results

    def retrieve(self, 
               corpus: List[Evidence], 
               queries: List[Question], 
               top_k: int, 
               return_sorted: bool = True, 
               chunk: bool = False,
               chunksize: int = 400000,
               **kwargs) -> Dict[str, Dict[str, float]]:
            print("chunk",chunk)
            if chunk:
                return self.retrieve_in_chunks(corpus,queries,top_k,return_sorted,chunksize=chunksize)

            corpus_texts = [(evidence.title() + self.sep + evidence.text()).strip() for evidence in corpus]
            queries_text = [question.text() for question in queries]
            with Run().context(RunConfig(nranks=1, experiment="default")):

                self.indexer.index(name="ambigqa", collection=corpus_texts, overwrite=True)

                self.indexer.get_index()
                searcher = Searcher(index="ambigqa", collection=corpus_texts)

                result_qrels = {}
                for idx,query in tqdm.tqdm(enumerate(queries)):
                    result_qrels[str(query.id())] = {}
                    results = searcher.search(query.text(), k=top_k)
                    for passage_id, passage_rank, passage_score in zip(*results):
                        result_qrels[str(query.id())][str(corpus[passage_id].id())] = passage_score
                #results = searcher.search_all(queries_text,k=100)
            return result_qrels


import tqdm
import time
from typing import List, Dict
from dexter.data.datastructures.evidence import Evidence
from dexter.data.datastructures.question import Question

from dexter.retriever.BaseRetriever import BaseRetriver
from dexter.retriever.lexical.ElasticSearch import ElasticSearch

def sleep(seconds):
    if seconds: time.sleep(seconds)
class BM25Search(BaseRetriver):
    def __init__(self, index_name: str, hostname: str = "https://localhost:9200",
                 cert_path:str="./http_cert",
                 elastic_passoword:str="password",
                 keys: Dict[str, str] = {"title": "title", "body": "text"}, 
                 language: str = "english",
                 batch_size: int = 128, timeout: int = 100, 
                 retry_on_timeout: bool = True, maxsize: int = 24, 
                 number_of_shards: int = "default", 
                 initialize: bool = True, sleep_for: int = 2):
        self.results = {}
        self.batch_size = batch_size
        self.initialize = initialize
        self.sleep_for = sleep_for
        self.config = {
            "hostname": hostname, 
            "index_name": index_name,
            "keys": keys,
            "timeout": timeout,
            "retry_on_timeout": retry_on_timeout,
            "maxsize": maxsize,
            "number_of_shards": number_of_shards,
            "language": language,
            "ca_certs":cert_path,
            "basic_auth":("elastic", elastic_passoword)

        }
        self.es = ElasticSearch(self.config)
        if self.initialize:
            self.initialise()
    
    def initialise(self):
        self.es.delete_index()
        sleep(self.sleep_for)
        self.es.create_index()
    
    def retrieve(self, corpus: List[Evidence], queries: List[Question], top_k: int, *args, **kwargs) -> Dict[str, Dict[str, float]]:
        
        corpus_final = {}
        for evidence in corpus:
            corpus_final[evidence.id()] = {"title":evidence.title(),"text":evidence.text()}
        # Index the corpus within elastic-search
        # False, if the corpus has been already indexed
        if self.initialize:
            self.index(corpus_final)
            # Sleep for few seconds so that elastic-search indexes the docs properly
            sleep(self.sleep_for)
        
        #retrieve results from BM25 
        query_ids = [query.id() for _, query in enumerate(queries)]
        # print("query_ids**",query_ids)
        queries = [query.text() for query in list(queries)]

        
        for start_idx in tqdm.trange(0, len(queries), self.batch_size, desc='progress'):
            query_ids_batch = query_ids[start_idx:start_idx+self.batch_size]
            results = self.es.lexical_multisearch(
                texts=queries[start_idx:start_idx+self.batch_size], 
                top_hits=top_k + 1) # Add 1 extra if query is present with documents
            
            for (query_id, hit) in zip(query_ids_batch, results):
                scores = {}
                for corpus_id, score in hit['hits']:
                        scores[corpus_id] = score
                self.results[str(query_id)] = scores
        
        return self.results
        
    
    def index(self, corpus: Dict[str, Dict[str, str]]):
        progress = tqdm.tqdm(unit="docs", total=len(corpus))
        # dictionary structure = {_id: {title_key: title, text_key: text}}
        dictionary = {idx: {
            self.config["keys"]["title"]: corpus[idx].get("title", None), 
            self.config["keys"]["body"]: corpus[idx].get("text", None)
            } for idx in list(corpus.keys())
        }
        self.es.bulk_add_to_index(
                generate_actions=self.es.generate_actions(
                dictionary=dictionary, update=False),
                progress=progress
                )